# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type-hinting utility for the package.
"""

__all__ = ["Typing"]


import types
import typing
from typing import NewType


try:
    Self = typing.Self
except AttributeError:  # pragma: no cover
    Self = NewType("Self", "Self")  # pragma: no cover
    # TODO @any: Remove when Python 3.11 is oldest supported version.


Cls = NewType("Cls", Self)


def _transpose_this_modules_types(
    class_dict: typing.Dict[str, typing.Any],
) -> None:
    """
    Inserts the types from this module's global namespace.
    """
    this_modules_types = {
        name: value
        for name, value in globals().items()
        if name[0].isupper()
    }
    class_dict.update(this_modules_types)


def _transpose_types_modules_types(
    class_dict: typing.Dict[str, typing.Any],
) -> None:
    """
    Inserts the types from the standard library's `types` module.
    """
    for name in types.__all__:
        if name[0].isupper():
            class_dict[name] = getattr(types, name)


def _transpose_typing_modules_types(
    class_dict: typing.Dict[str, typing.Any],
) -> None:
    """
    Inserts the types from the standard library's `typing` module.
    """
    for name in typing.__all__:
        if name[0].isupper():
            class_dict[name] = getattr(typing, name)


class Typing:
    """
    A container for type-hinting variables.
    """

    __slots__ = ()

    _transpose_this_modules_types(class_dict=vars())
    _transpose_types_modules_types(class_dict=vars())
    _transpose_typing_modules_types(class_dict=vars())

    overload = typing.overload
    runtime_checkable = typing.runtime_checkable

    @classmethod
    def _test_type(cls, new_type: type) -> None:
        """
        Throws `TypeError` if `new_type` doesn't have class-type
        attributes.
        """
        has_type_attributes = (
            hasattr(new_type, "mro")
            and hasattr(new_type, "__mro__")
            and hasattr(new_type, "__bases__")
            and hasattr(new_type, "__prepare__")
        )
        if not has_type_attributes:
            raise TypeError(f"{new_type!r} is not a type.")

    @classmethod
    def _test_type_name(cls, name: str) -> None:
        """
        Assures new type additions to the class are unique & title or
        capital-cased identifiers.
        """
        attribute_already_defined = name in cls.__dict__
        is_mixed_case = "_" in name
        is_capitalized = name[0].isupper()

        if not name.isidentifier():
            raise ValueError(f"Invalid type name {name!r}.")
        elif attribute_already_defined:
            raise AttributeError(f"{name!r} is already defined.")
        elif is_mixed_case or not is_capitalized:
            raise ValueError(f"{name!r} must be title or capital-cased")

    @classmethod
    def add_type(cls, new_type: type) -> type:
        """
        Adds a new typing type to the class dictionary.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        class MessageType(bytes):
            pass

        Typing.add_type(MessageType)
        message: Typing.MessageType = b"Hello, World!"
        """
        name = new_type.__qualname__
        cls._test_type(new_type)
        cls._test_type_name(name)
        setattr(cls, name, new_type)
        return new_type


module_api = dict(
    Typing=Typing,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface to X25519 protocols.
"""

__all__ = ["X25519"]


from aiootp._typing import Typing as t
from aiootp.asynchs import asleep
from aiootp.keygens.domain_kdf import DomainKDF

from .shared_interface import Base25519
from .double_diffie_hellman_client import DoubleDiffieHellmanClient
from .double_diffie_hellman_server import DoubleDiffieHellmanServer
from .triple_diffie_hellman_client import TripleDiffieHellmanClient
from .triple_diffie_hellman_server import TripleDiffieHellmanServer


class X25519(Base25519):
    """
    This class is used to create stateful objects that simplify usage of
    the cryptography library's x25519 protocol.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    user_alice = X25519().generate()
    # Alice wants to create a shared key with Bob. So, Alice sends the
    # public key bytes of her new key to bob ->
    internet.send(user_alice.public_bytes)

    # In a land far away ->
    alices_message = internet.receive()

    # Bob sees the message from Alice! So she creates a key to accept
    # the exchange & sends the public bytes back to Alice ->
    user_bob = await X25519().agenerate()
    shared_key = user_bob.exchange(alices_message)
    internet.send(user_bob.public_bytes)

    # When Alice receives Bob's public key & finishes the exchange, they
    # will have a shared symmetric key to encrypt messages to one
    # another.
    bobs_response = internet.receive()
    shared_key = user_alice.exchange(bobs_response)

    This protocol is not secure against active adversaries that can
    manipulate the information while its in transit between Alice &
    Bob. Each public key should only be used once.
    """

    __slots__ = ("_public_key", "_secret_key")

    _KDF: type = DomainKDF
    _DoubleDiffieHellmanClient: type = DoubleDiffieHellmanClient
    _DoubleDiffieHellmanServer: type = DoubleDiffieHellmanServer
    _TripleDiffieHellmanClient: type = TripleDiffieHellmanClient
    _TripleDiffieHellmanServer: type = TripleDiffieHellmanServer

    PublicKey: type = t.X25519PublicKey
    SecretKey: type = t.X25519PrivateKey

    async def aexchange(
        self, public_key: t.Union[bytes, t.PublicKeyType], /
    ) -> bytes:
        """
        Takes in a public key from a communicating party & uses the
        instance's secret key to do an elliptic curve diffie-hellman
        exchange & returns the resulting secret shared bytes.
        """
        await asleep()
        if public_key.__class__ is bytes:
            public_key = self.PublicKey.from_public_bytes(public_key)
        return self._secret_key.exchange(public_key)

    def exchange(
        self, public_key: t.Union[bytes, t.PublicKeyType], /
    ) -> bytes:
        """
        Takes in a public key from a communicating party & uses the
        instance's secret key to do an elliptic curve diffie-hellman
        exchange & returns the resulting secret shared bytes.
        """
        if public_key.__class__ is bytes:
            public_key = self.PublicKey.from_public_bytes(public_key)
        return self._secret_key.exchange(public_key)

    @classmethod
    def dh2_client(cls, /) -> DoubleDiffieHellmanClient:
        """
        Uses a `X25519` ephemeral secret key & a peer's identity &
        ephemeral public keys to enact the client side of a 2DH deniable
        key exchange.

         _____________________________________
        |                                     |
        |          Protocol Diagram:          |
        |_____________________________________|

                -----------------          |         -----------------
                |  Client-side  |          |         |  Server-side  |
                -----------------          |         -----------------
                                           |
                                           |         X25519().generate() = key
                                           |
        client = X25519.dh2_client()       |           key.public_bytes = id_s
                                           |
        eph_c = client.send(id_s)       ------>
                                           |
                                           |         key.dh2_server() = server
                                           |
                                           |       server.receive(eph_c) = kdf
                                           |
                                        <------          server.send() = eph_s
                                           |
        kdf = client.receive(eph_s)        |
                                           |
        ------------------------------------------------------------------------
        """
        return cls._DoubleDiffieHellmanClient(cls, kdf_type=cls._KDF)

    def dh2_server(self, /) -> DoubleDiffieHellmanServer:
        """
        Uses `X25519` identity & ephemeral secret keys, & a peer's
        ephemeral public key to enact the server side of a 2DH deniable
        key exchange.

         _____________________________________
        |                                     |
        |          Protocol Diagram:          |
        |_____________________________________|

                -----------------          |         -----------------
                |  Client-side  |          |         |  Server-side  |
                -----------------          |         -----------------
                                           |
                                           |         X25519().generate() = key
                                           |
        client = X25519.dh2_client()       |           key.public_bytes = id_s
                                           |
        eph_c = client.send(id_s)       ------>
                                           |
                                           |         key.dh2_server() = server
                                           |
                                           |       server.receive(eph_c) = kdf
                                           |
                                        <------          server.send() = eph_s
                                           |
        kdf = client.receive(eph_s)        |
                                           |
        ------------------------------------------------------------------------
        """
        return self._DoubleDiffieHellmanServer(self, kdf_type=self._KDF)

    def dh3_client(self, /) -> TripleDiffieHellmanClient:
        """
        Uses `X25519` identity & ephemeral secret keys, & a peer's
        identity & ephemeral public keys to enact the client side of a
        3DH deniable key exchange.

         _____________________________________
        |                                     |
        |          Protocol Diagram:          |
        |_____________________________________|

                -----------------          |         -----------------
                |  Client-side  |          |         |  Server-side  |
                -----------------          |         -----------------
                                           |
        key = X25519().generate()          |         X25519().generate() = key
                                           |
        client = key.dh3_client()          |           key.public_bytes = id_s
                                           |
        id_c, eph_c = client.send(id_s) ------>
                                           |
                                           |         key.dh3_server() = server
                                           |
                                           | server.receive(id_c, eph_c) = kdf
                                           |
                                        <------          server.send() = eph_s
                                           |
        kdf = client.receive(eph_s)        |
                                           |
        ------------------------------------------------------------------------
        """
        return self._TripleDiffieHellmanClient(self, kdf_type=self._KDF)

    def dh3_server(self, /) -> TripleDiffieHellmanServer:
        """
        Uses `X25519` identity & ephemeral secret keys, & a peer's
        identity & ephemeral public keys to enact the server side of a
        3DH deniable key exchange.

         _____________________________________
        |                                     |
        |          Protocol Diagram:          |
        |_____________________________________|

                -----------------          |         -----------------
                |  Client-side  |          |         |  Server-side  |
                -----------------          |         -----------------
                                           |
        key = X25519().generate()          |         X25519().generate() = key
                                           |
        client = key.dh3_client()          |           key.public_bytes = id_s
                                           |
        id_c, eph_c = client.send(id_s) ------>
                                           |
                                           |         key.dh3_server() = server
                                           |
                                           | server.receive(id_c, eph_c) = kdf
                                           |
                                        <------          server.send() = eph_s
                                           |
        kdf = client.receive(eph_s)        |
                                           |
        ------------------------------------------------------------------------
        """
        return self._TripleDiffieHellmanServer(self, kdf_type=self._KDF)


module_api = dict(
    X25519=t.add_type(X25519),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import warnings
from hashlib import shake_128
from collections import deque

from conftest import *

from aiootp.asynchs.clocks import s_counter
from aiootp.randoms.entropy_daemon import EntropyDaemon
from aiootp.randoms.simple import arandom_sleep, random_sleep
from aiootp.randoms.threading_safe_entropy_pool import (
    ThreadingSafeEntropyPool,
)
from aiootp.randoms.rng import (
    arandom_number_generator,
    random_number_generator,
)


class TestRandomSleeps:
    runs = 32
    span = 0.001
    max_overhead = 0.02
    max_ideal_overhead = 0.001

    async def test_async_random_sleep(self) -> None:
        MAX = (0.6 * self.span) + self.max_overhead
        total = 0
        for _ in range(self.runs):
            start = s_counter()
            await arandom_sleep(self.span)
            end = s_counter()
            total += end - start

        average_arandom_sleep = total / self.runs
        assert MAX >= average_arandom_sleep >= 0
        if average_arandom_sleep >= (
            0.6 * self.span + self.max_ideal_overhead
        ):
            warnings.warn(
                f"NOTICE: span={self.span} : {average_arandom_sleep=}"
            )

    async def test_sync_random_sleep(self) -> None:
        MAX = (0.6 * self.span) + self.max_overhead
        total = 0
        for _ in range(self.runs):
            start = s_counter()
            random_sleep(self.span)
            end = s_counter()
            total += end - start

        average_random_sleep = total / self.runs
        assert MAX >= average_random_sleep >= 0
        if average_random_sleep >= (
            0.6 * self.span + self.max_ideal_overhead
        ):
            warnings.warn(
                f"NOTICE: span={self.span} : {average_random_sleep=}"
            )


class TestThreadingSafeEntropyPool:
    cls = ThreadingSafeEntropyPool

    async def test_hasher_methods(self) -> None:
        token = token_bytes(32)
        obj = self.cls(token, obj=shake_128, pool=[token])
        hasher = obj._obj.copy()
        obj_copy = obj.copy()
        assert obj.name == hasher.name
        assert obj.name == obj_copy.name
        assert obj.block_size == hasher.block_size
        assert obj.block_size == obj_copy.block_size
        assert obj.digest_size == hasher.digest_size
        assert obj.digest_size == obj_copy.digest_size
        assert obj.digest(32) == hasher.digest(32)
        assert obj.digest(32) == obj_copy.digest(32)
        assert obj.hexdigest(32) == hasher.hexdigest(32)
        assert obj.hexdigest(32) == obj_copy.hexdigest(32)

        obj.update(token)
        obj_copy.update(token)
        hasher.update(token)
        assert obj.digest(32) == hasher.digest(32)
        assert obj.digest(32) == obj_copy.digest(32)
        assert obj.hexdigest(32) == hasher.hexdigest(32)
        assert obj.hexdigest(32) == obj_copy.hexdigest(32)


class TestEntropyDaemon:
    pool = deque([csprng(32), csprng(32)], maxlen=32)
    gadget = ThreadingSafeEntropyPool(csprng(), obj=shake_128, pool=pool)
    daemon = EntropyDaemon(
        entropy_pool=pool, gadget=gadget, max_delay=0.025
    )

    async def test_aaa_start_daemon(self) -> None:
        assert not hasattr(self.daemon, "_daemon")
        self.daemon.start()
        assert hasattr(self.daemon, "_daemon")
        assert self.daemon._daemon.is_alive()

    async def test_max_delay_limits(self) -> None:
        problem = (  # fmt: skip
            "A negative delay was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            self.daemon.set_max_delay(-1)

    async def test_temporary_max_delay_limits(self) -> None:
        problem = (  # fmt: skip
            "A negative temporary delay was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            self.daemon.set_temporary_max_delay(-1)

    async def test_temporary_max_delay_reverts_after_duration(self) -> None:
        duration = 0.001
        initial_delay = self.daemon._max_delay
        new_delay = initial_delay / 2
        self.daemon.set_temporary_max_delay(new_delay, duration=duration)

        assert new_delay == self.daemon._max_delay
        max_wait_time = initial_delay + new_delay + duration
        await asleep(max_wait_time)
        assert initial_delay == self.daemon._max_delay

    async def test_zzz_cancel_stops_daemon(self) -> None:
        assert self.daemon._daemon.is_alive()
        self.daemon.cancel()
        await asleep(3 * self.daemon._max_delay)
        assert not self.daemon._daemon.is_alive()


class TestCSPRNG:
    kw = dict(entropy=token_bytes(16))

    async def test_declared_output_sizes_are_respected(self) -> None:
        for size in range(8, 257, 32):
            result = await acsprng(size, **self.kw)
            assert size == len(result)

            result = csprng(size, **self.kw)
            assert size == len(result)

    async def test_userland_entropy_can_be_any_type(self) -> None:
        for datum in (
            token_bytes(32).hex(),
            token_bits(32),
            [None, "test"],
        ):
            await acsprng(entropy=datum)
            csprng(entropy=datum)

    async def test_async_thread_safe_entropy(self) -> None:
        async def try_to_make_duplicate_readouts() -> None:
            for _ in range(runs):
                entropy_pool.add(await acsprng())

        runs = 32
        entropy_pool = set()
        await Threads.agather(
            *[try_to_make_duplicate_readouts for _ in range(runs)]
        )
        assert runs**2 == len(entropy_pool)

    async def test_sync_thread_safe_entropy(self) -> None:
        def try_to_make_duplicate_readouts() -> None:
            for _ in range(runs):
                entropy_pool.add(csprng())

        runs = 32
        entropy_pool = set()
        Threads.gather(
            *[try_to_make_duplicate_readouts for _ in range(runs)]
        )
        assert runs**2 == len(entropy_pool)


class TestRandomNumberGenerator:
    kw = dict(freshness=0, entropy=token_bytes(16))

    async def test_freshness_must_be_positive_int(self) -> None:
        problem = (  # fmt: skip
            "A non-positive int `freshness` value was allowed."
        )
        for bad_freshness in (b"-2", "1.0"):
            with Ignore(TypeError, if_else=violation(problem)):
                await arandom_number_generator(freshness=bad_freshness)
        for bad_freshness in (-1, False, 1.0):
            with Ignore(ValueError, if_else=violation(problem)):
                await arandom_number_generator(freshness=bad_freshness)

    async def test_async_declared_output_sizes_are_respected(self) -> None:
        for size in range(8, 257, 32):
            result = await arandom_number_generator(size, **self.kw)
            assert size == len(result)

    def test_sync_declared_output_sizes_are_respected(self) -> None:
        for size in range(8, 257, 32):
            result = random_number_generator(size, **self.kw)
            assert size == len(result)

    async def test_async_userland_entropy_can_be_any_type(self) -> None:
        for datum in (
            token_bytes(32).hex(),
            token_bits(32),
            [None, "test"],
        ):
            await arandom_number_generator(freshness=0, entropy=datum)

    def test_sync_userland_entropy_can_be_any_type(self) -> None:
        for datum in (
            token_bytes(32).hex(),
            token_bits(32),
            [None, "test"],
        ):
            random_number_generator(freshness=0, entropy=datum)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type for creating unblinded unique bytes values.
"""

__all__ = ["RawGUID"]


from aiootp._typing import Typing as t
from aiootp._constants import BIG
from aiootp._exceptions import Issue
from aiootp.commons import ConfigMap, FrozenInstance
from aiootp.asynchs import asleep

from .raw_guid_config import RawGUIDConfig, RawGUIDContainer


class RawGUID(FrozenInstance):
    """
    Creates a unique bytes value by combining a timestamp, random bytes,
    an instance-specific node-ID, & an incrementing counter.

     _____________________________________
    |                                     |
    |   Format Diagram: Default Raw GUID  |
    |_____________________________________|
     __________________________________________________________________
    |                         |                    |         |         |
    |      timestamp (ns)     |       token        | node-id |  ticker |
    |-------------------------|--------------------|---------|---------|
    |         8-bytes         |       6-bytes      |  1-byte |  1-byte |
    |_________________________|____________________|_________|_________|
    |                                                                  |
    |                          16-bytes                                |
    |__________________________________________________________________|
    """

    __slots__ = ("tick", "node_id", "config")

    _configs = ConfigMap(
        mapping={
            config_id: RawGUIDConfig(
                timestamp_bytes=8,
                prf_bytes=config_id - 10,
                node_id_bytes=1,
                ticker_bytes=1,
                size=config_id,
            )
            for config_id in range(12, 33)
        },
        config_type=RawGUIDConfig,
    )

    def _process_node_id(self, node_id: bytes) -> bytes:
        """
        Raises `ValueError` if the specified `node_id` isn't the correct
        length.
        """
        node_id_bytes = self.config.NODE_ID_BYTES
        if len(node_id) != node_id_bytes:
            raise Issue.invalid_length("node_id", node_id_bytes)
        return node_id

    def __init__(
        self, node_id: bytes = b"\x00", *, config_id: t.Hashable = 16
    ) -> None:
        self.config = self._configs[config_id]
        self.node_id = self._process_node_id(node_id)
        self.tick = 0

    def _incremented_tick(self) -> bytes:
        """
        Returns the incremented counter value.
        """
        c = self.config
        object.__setattr__(self, "tick", (self.tick + 1) & c.TICK_DOMAIN)
        return self.tick.to_bytes(c.TICKER_BYTES, BIG)

    async def anew(self) -> bytes:
        """
        Creates an unblinded, raw GUID bytes value.
        """
        await asleep()
        return self.new()

    def new(self) -> bytes:
        """
        Creates an unblinded, raw GUID bytes value.
        """
        config = self.config
        return (
            config.clock.make_timestamp(size=config.TIMESTAMP_BYTES)
            + config.prf(config.PRF_BYTES)
            + self.node_id
            + self._incremented_tick()
        )

    async def aread(self, guid: bytes) -> RawGUIDContainer:
        """
        Parses unblinded, unique bytes values into a mapping.
        """
        await asleep()
        return self.read(guid)

    def read(self, guid: bytes) -> RawGUIDContainer:
        """
        Parses unblinded, unique bytes values into a mapping.
        """
        return RawGUIDContainer(guid, config=self.config)


module_api = dict(
    RawGUID=t.add_type(RawGUID),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Functions to transform the representation of data.
"""

__all__ = [
    "abase_as_int",
    "aint_as_base",
    "axi_mix",
    "base_as_int",
    "int_as_base",
    "xi_mix",
]


from aiootp._typing import Typing as t
from aiootp._constants import BIG, Tables
from aiootp._exceptions import Issue
from aiootp._gentools import abatch, batch
from aiootp.asynchs import asleep


async def abase_as_int(
    string: t.AnyStr,
    base: t.Optional[int] = None,
    *,
    table: t.AnyStr = Tables.ASCII_95,
) -> int:
    """
    Convert `string` in numerical `base` into decimal integer.
    """
    if base is None:
        base = len(table)
    power = 1
    result = 0
    base_table = table[:base]
    await asleep()
    for char in reversed(string):
        if char not in base_table:
            raise Issue.invalid_value("base with the given table")
        result += base_table.find(char) * power
        power = power * base
    await asleep()
    return result


def base_as_int(
    string: t.AnyStr,
    base: t.Optional[int] = None,
    *,
    table: t.AnyStr = Tables.ASCII_95,
) -> int:
    """
    Convert `string` in numerical `base` into decimal integer.
    """
    if base is None:
        base = len(table)
    power = 1
    result = 0
    base_table = table[:base]
    for char in reversed(string):
        if char not in base_table:
            raise Issue.invalid_value("base with the given table")
        result += base_table.find(char) * power
        power = power * base
    return result


async def aint_as_base(
    number: int,
    base: t.Optional[int] = None,
    *,
    table: t.AnyStr = Tables.ASCII_95,
) -> t.AnyStr:
    """
    Convert an `number` back into a string in numerical `base`.
    """
    if not number:
        return table[:1]
    elif base is None:
        base = len(table)
    digits = []
    base_table = table[:base]
    await asleep()
    while number:
        digits.append(base_table[number % base])
        number //= base
    digits.reverse()
    await asleep()
    if base_table.__class__ is bytes:
        return bytes(digits)
    else:
        return "".join(digits)


def int_as_base(
    number: int,
    base: t.Optional[int] = None,
    *,
    table: t.AnyStr = Tables.ASCII_95,
) -> t.AnyStr:
    """
    Convert an `number` back into a string in numerical `base`.
    """
    if not number:
        return table[:1]
    elif base is None:
        base = len(table)
    digits = []
    base_table = table[:base]
    while number:
        digits.append(base_table[number % base])
        number //= base
    digits.reverse()
    if base_table.__class__ is bytes:
        return bytes(digits)
    else:
        return "".join(digits)


async def axi_mix(bytes_hash: bytes, size: int = 8) -> bytes:
    """
    Xors subsequent `size` length segments of `bytes_hash` with each
    other to condense the hash down to `size` bytes.
    """
    result = 0
    async for chunk in abatch(bytes_hash, size=size):
        result ^= int.from_bytes(chunk, BIG)
    return result.to_bytes(size, BIG)


def xi_mix(bytes_hash: bytes, size: int = 8) -> bytes:
    """
    Xors subsequent `size` length segments of `bytes_hash` with each
    other to condense the hash down to `size` bytes.
    """
    result = 0
    for chunk in batch(bytes_hash, size=size):
        result ^= int.from_bytes(chunk, BIG)
    return result.to_bytes(size, BIG)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    abase_as_int=abase_as_int,
    aint_as_base=aint_as_base,
    axi_mix=axi_mix,
    base_as_int=base_as_int,
    int_as_base=int_as_base,
    xi_mix=xi_mix,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A collection constants & precomputed values to standardize, speed
up, & clean up their usage in the package.
"""

import typing
import hashlib
from hashlib import sha3_256, sha3_512, shake_128, shake_256

from aiootp._typing import Typing as t

from . import NamespaceMapping, collect_non_private_constants


INT_BYTES: int = 8
B_TO_MB_RATIO: int = 1024 * 1024

BYTES_FLAG: bytes = b"\x01\x02"
BYTES_FLAG_SIZE: int = 2

FILENAME_HASH_BYTES: int = 24

EPOCH: int = 1672531200  # Sun, 01 Jan 2023 00:00:00 UTC
SAFE_TIMESTAMP_BYTES: int = 8
DEFAULT_TTL: typing.Optional[int] = None
DEFAULT_TIMEOUT: int = 0
NS_TO_S_RATIO: int = 1_000_000_000


OMITTED: str = "<omitted>"

TB_PORT: int = 9150
TOR_PORT: int = 9050

# fmt: off
_algorithms_available: t.Set[str] = set(hashlib.algorithms_available)
HASHER_TYPES: NamespaceMapping = NamespaceMapping(**{
    str(name): NamespaceMapping(
        factory=factory,
        digest_size=factory().digest_size,
        block_size=factory().block_size,
    )
    for name, factory in hashlib.__dict__.items()
    if name in _algorithms_available
})
SHA3_STATE_SIZE: int = 200
SHA3_256_BLOCKSIZE: int = sha3_256().block_size
SHA3_512_BLOCKSIZE: int = sha3_512().block_size
SHAKE_128_BLOCKSIZE: int = shake_128().block_size
SHAKE_256_BLOCKSIZE: int = shake_256().block_size
# fmt: on

MIN_KEY_BYTES: int = 64
DEFAULT_AAD: bytes = b""


# String Names:
ACTIVE: str = "active_connection"
ADDRESS: str = "address"
ADMIN: str = "admin"
AAD: str = "aad"
AGE: str = "age_of_connection"
AIOOTP: str = "aiootp"
ASYNC: str = "asynchronous"
AUTHENTICATED_ASSOCIATED_DATA: str = "authenticated_associated_data"
AUTHENTICATION: str = "authentication"
AUTHENTICATION_KEY: str = "authentication_key"
BIG: str = "big"
BLOCK_ID: str = "block_id"
CHANNEL: str = "channel"
CHANNELS: str = "channels"
CHECKSUM: str = "checksum"
CHECKSUMS: str = "checksums"
CHUNKY_2048: str = "Chunky2048"
CIPHERTEXT: str = "ciphertext"
CLIENT: str = "client"
CLIENT_ADDRESS: str = "client_address"
CLIENT_ID: str = "client_id"
CLIENT_INDEX: str = "client_database_index"
CLIENT_KEY: str = "client_key"
CLIENT_MESSAGE_KEY: str = "client_message_key"
CONFIG: str = "config"
CONFIG_ID: str = "CONFIG_ID"
CONFIG_TYPE: str = "CONFIG_TYPE"
CONTROL_BITS: str = "control_bits"
CONVERSATION: str = "conversation"
CORES: str = "cores"
CORRUPT: str = "corrupt"
CPU: str = "cpu"
CSPRNG: str = "cryptographically_secure_prng"
DATABASE: str = "database"
DAY: str = "day"
DAYS: str = "days"
DECRYPT: str = "decrypt"
DECRYPTION: str = "decryption"
DH2: str = "diffie_hellman_2x"
DH3: str = "diffie_hellman_3x"
DIFFIE_HELLMAN: str = "diffie_hellman"
DIRECTORY: str = "directory"
ECDHE: str = "elliptic_curve_diffie_hellman"
EMAIL_ADDRESS: str = "email_address"
ENCRYPT: str = "encrypt"
ENCRYPTION: str = "encryption"
ENCRYPTION_KEY: str = "encryption_key"
ENTROPY: str = "entropy"
EPHEMERAL_KEY: str = "ephemeral_key"
EQUALITY: str = "equality"
EXTENDED_DH_EXCHANGE: str = "extended_diffie_hellman_exchange"
FAILED: str = "failed"
FILE_KEY: str = "file_key"
FILENAME: str = "filename"
GUEST: str = "guest"
GUID_CLOCK_MASK: str = "guid_clock_mask"
GUID_SALT: str = "guid_salt"
HEADER: str = "header"
HMAC: str = "hmac"
HOUR: str = "hour"
HOURS: str = "hours"
HTTP: str = "http"
HTTPS: str = "https"
ID: str = "contact_id"
IDENTITY_KEY: str = "identity_key"
INACTIVE: str = "terminated_connection"
INNER_HEADER: str = "inner_header"
IV: str = "iv"
KEY: str = "key"
KDF: str = "key_derivation_function"
KEEP_ALIVE: str = "keep_alive"
KEY_ID: str = "key_id"
KEYSTREAM: str = "keystream"
LISTENING: str = "listening"
LITTLE: str = "little"
MAINTAINING: str = "maintaining"
MANIFEST: str = "manifest"
MANUAL: str = "manual_mode"
MAX_INACTIVITY: str = "max_inactivity"
MB: str = "mb"
MESSAGE_ID: str = "message_id"
MESSAGE_KEY: str = "message_key"
MESSAGE_NUMBER: str = "message_number"
MESSAGE: str = "message"
MESSAGES: str = "messages"
METADATA: str = "metadata"
METATAG: str = "metatag"
METATAG_KEY: str = "metatag_key"
MICROSECOND: str = "microsecond"
MICROSECONDS: str = "microseconds"
MILLISECOND: str = "millisecond"
MILLISECONDS: str = "milliseconds"
MINUTE: str = "minute"
MINUTES: str = "minutes"
MNEMONIC: str = "mnemonic"
MONTH: str = "month"
MONTHS: str = "months"
NANOSECOND: str = "nanosecond"
NANOSECONDS: str = "nanoseconds"
NEW_CONTACT: str = "new_contact"
OLD_KEY: str = "last_shared_key"
ONION: str = "onion"
PACKAGE_SIGNER: str = "PackageSigner"
PASSCRYPT: str = "passcrypt"
PASSPHRASE: str = "passphrase"
PAYLOAD: str = "payload"
PERIOD_KEY: str = "period_key"
PERIOD_KEYS: str = "period_keys"
PERMUTATION: str = "permutation"
PERMUTATION_KEY: str = "permutation_key"
PHASE: str = "phase"
PHONE_NUMBER: str = "phone_number"
PLAINTEXT: str = "plaintext"
PRNG: str = "pseudo_random_number_generator"
PUBLIC_CREDENTIALS: str = "public_credentials"
PUBLIC_KEY: str = "public_key"
RACHET: str = "rachet_shared_key"
RECEIVING: str = "receiving"
RECEIVING_COUNT: str = "receiving_count"
RECEIVING_KEYS: str = "receiving_keys"
RECEIVING_STREAM: str = "receiving_stream"
REGISTRATION: str = "registration"
RETRY: str = "retry"
SALT: str = "salt"
SALT_SIZE: str = "salt_size"
SCHEMA: str = "schema"
SCOPE: str = "scope"
SECOND: str = "second"
SECONDS: str = "seconds"
SECRET: str = "secret"
SECRET_CREDENTIALS: str = "secret_credentials"
SECRET_KEY: str = "secret_key"
SECURE_CHANNEL: str = "secure_channel"
SEED: str = "seed"
SENDER: str = "sender"
SENDING: str = "sending"
SENDING_COUNT: str = "sending_count"
SENDING_KEYS: str = "sending_keys"
SENDING_STREAM: str = "sending_stream"
SERVER: str = "server"
SERVER_ADDRESS: str = "server_address"
SERVER_ID: str = "server_id"
SERVER_INDEX: str = "server_database_index"
SERVER_KEY: str = "server_key"
SERVER_MESSAGE_KEY: str = "server_message_key"
SESSION_ID: str = "session_id"
SESSION_KEY: str = "session_key"
SESSION_TOKEN: str = "session_tracking_token"
SHARED_KEY: str = "shared_key"
SHARED_KEYS: str = "shared_keys"
SHARED_SECRET: str = "shared_secret"
SHARED_SEED: str = "shared_seed"
SHMAC: str = "shmac"
SIGNAL: str = "signal"
SIGNALS: str = "signals"
SIGNATURE: str = "signature"
SIGNING_KEY: str = "signing_key"
SIV: str = "synthetic_iv"
SIV_KEY: str = "siv_key"
SLICK_256: str = "Slick256"
STATE: str = "state"
STATUS: str = "status"
SUCCESS: str = "success"
SYNC: str = "synchronous"
TAG: str = "tag"
TAG_SIZE: str = "tag_size"
THREAD_SAFE_ENTROPY: str = "thread_safe_entropy"
TIMEOUT: str = "timeout"
TIMESTAMP: str = "timestamp"
TOKEN: str = "token"
TTL: str = "ttl"
UNSENT_MESSAGES: str = "unsent_messages"
URL: str = "url"
USER: str = "user"
USERNAME: str = "username"
UUID: str = "unique_user_id"
VERIFICATION: str = "verification"
VERSIONS: str = "versions"
YEAR: str = "year"
YEARS: str = "years"


# Dynamically Generated:
FILENAME_HASH_SLICE: slice = slice(None, FILENAME_HASH_BYTES)

EPOCH_NS: int = 1_000_000_000 * EPOCH


# Create a binding to the module's definitions for UI
definitions = collect_non_private_constants(globals())


__all__ = [*definitions]


module_api = dict(
    **definitions,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Domain separation constants & utilities.
"""

__all__ = ["Domains", "DomainEncoder"]


from hashlib import shake_128

from aiootp._typing import Typing as t
from aiootp._constants import misc, DEFAULT_AAD
from aiootp.commons import FrozenInstance

from .hashing import ahash_bytes, hash_bytes


class DomainEncoder(FrozenInstance):
    """
    A base class which enables domain constants to be encoded & created
    for specific use cases.
    """

    __slots__ = ()

    _DOMAIN: bytes = b"domain_constant_encoder"

    @classmethod
    async def aencode_constant(
        cls,
        constant: t.AnyStr,
        size: int = 8,
        *,
        domain: bytes = b"",
        aad: bytes = DEFAULT_AAD,
        pad: bytes = b"\x00",
    ) -> bytes:
        """
        Receives a `str` or `bytes`-type `constant`, encodes & hashes
        it under a `domain`, along with the metadata of the encoding,
        then returns the `size`-byte digest from the `shake_128` XOF.

        These returned values are used by the package as inputs to other
        functions which in turn makes their outputs domain-specific.
        This has various security benefits, such as:

        https://eprint.iacr.org/2010/264.pdf & more recent published
        works show schemes which are not provably secure, may be
        transformable into provably secure schemes just with some
        assumptions that functions which they rely upon happen to be
        domain-specific.
        """
        if constant.__class__ is not bytes:
            constant = constant.encode()
        return await ahash_bytes(
            cls._DOMAIN,
            domain,
            aad,
            constant,
            hasher=shake_128,
            size=size,
            pad=pad,
        )

    @classmethod
    def encode_constant(
        cls,
        constant: t.AnyStr,
        size: int = 8,
        *,
        domain: bytes = b"",
        aad: bytes = DEFAULT_AAD,
        pad: bytes = b"\x00",
    ) -> bytes:
        """
        Receives a `str` or `bytes`-type `constant`, encodes & hashes
        it under a `domain`, along with the metadata of the encoding,
        then returns the `size`-byte digest from the `shake_128` XOF.

        These returned values are used by the package as inputs to other
        functions which in turn makes their outputs domain-specific.
        This has various security benefits, such as:

        https://eprint.iacr.org/2010/264.pdf & more recent published
        works show schemes which are not provably secure, may be
        transformable into provably secure schemes just with some
        assumptions that functions which they rely upon happen to be
        domain-specific.
        """
        if constant.__class__ is not bytes:
            constant = constant.encode()
        return hash_bytes(
            cls._DOMAIN,
            domain,
            aad,
            constant,
            hasher=shake_128,
            size=size,
            pad=pad,
        )


class Domains(DomainEncoder):
    """
    A collection of encoded constants which can augment function inputs
    to make their outputs domain-specific.
    """

    __slots__ = ()

    for name, value in misc.__dict__.items():
        if value.__class__ is str:
            vars()[name] = DomainEncoder.encode_constant(value)

    del name
    del value


module_api = dict(
    DomainEncoder=t.add_type(DomainEncoder),
    Domains=t.add_type(Domains),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Randomness generation, tools, & unique pseudo-random identifiers.
"""

__all__ = ["GUID", "SequenceID", "acsprng", "csprng"]


from .simple import *
from .threading_safe_entropy_pool import *
from .rng import *
from .entropy_daemon import *
from .ids import *


subpackages = dict(ids=ids)


modules = dict(
    _early_salts=_early_salts,
    entropy_daemon=entropy_daemon,
    rng=rng,
    simple=simple,
    threading_safe_entropy_pool=threading_safe_entropy_pool,
)


module_api = dict(
    GUID=GUID,
    SequenceID=SequenceID,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    achoice=achoice,
    acsprng=acsprng,
    arandom_number_generator=arandom_number_generator,
    arandom_sleep=arandom_sleep,
    atoken_bits=atoken_bits,
    atoken_bytes=atoken_bytes,
    auniform=auniform,
    choice=choice,
    csprng=csprng,
    random_number_generator=random_number_generator,
    random_sleep=random_sleep,
    token_bits=token_bits,
    token_bytes=token_bytes,
    uniform=uniform,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Lower assurance randomness functions.
"""

__all__ = [
    "acanonical_token",
    "achoice",
    "arandom_sleep",
    "atoken_bits",
    "atoken_bytes",
    "auniform",
    "canonical_token",
    "choice",
    "random_sleep",
    "token_bits",
    "token_bytes",
    "uniform",
]


import random as _stdlib_random
from secrets import choice, token_bytes
from secrets import randbits as token_bits

from aiootp._typing import Typing as t
from aiootp._constants import BIG
from aiootp.asynchs import get_process_id, get_thread_id
from aiootp.asynchs import asleep, sleep, ns_counter


# initializing weakly entropic functionalities
_random = _stdlib_random.Random(token_bytes(2500))
uniform = _random.uniform


async def auniform(a: t.RealNumber, b: t.RealNumber) -> float:
    """
    Asynchronous version of the standard library's `random.uniform`.
    """
    await asleep()
    return uniform(a=a, b=b)


async def achoice(seq: t.Sequence[t.Any]) -> t.Any:
    """
    Asynchronous version of the standard library's `secrets.choice`.
    """
    await asleep()
    return choice(seq=seq)


async def arandom_sleep(seconds: t.PositiveRealNumber = 1) -> t.Any:
    """
    Asynchronously sleeps for a psuedo-random duration less than the
    provided maximum number of `seconds`.
    """
    await asleep(seconds * await auniform(0, 1))


def random_sleep(seconds: t.PositiveRealNumber = 1) -> t.Any:
    """
    Synchronously sleeps for a psuedo-random duration less than the
    provided maximum number of `seconds`.
    """
    sleep(seconds * uniform(0, 1))


async def atoken_bits(k: int) -> int:
    """
    Returns `k` entropic bits from `secrets.randbits`.
    """
    await asleep()
    return token_bits(k=k)


async def atoken_bytes(nbytes: int) -> bytes:
    """
    Returns `nbytes` entropic bytes from `secrets.token_bytes`.
    """
    await asleep()
    return token_bytes(nbytes=nbytes)


async def acanonical_token() -> bytes:
    """
    Combines 256-bits of system entropy, a monotonic nanosecond counter,
    as well as process & thread identifiers into a 56-byte token to
    uniquely identify the current context of the function call. These
    tokens SHOULD NOT be exposed publicly.
     _____________________________________
    |                                     |
    |           Format Diagram:           |
    |_____________________________________|
     ___________________________________________________________________
    |                            |            |            |            |
    |         token_bits         |  time (ns) |  thread ID | process ID |
    |----------------------------|------------|------------|------------|
    |                            |            |            |            |
    |          32-bytes          |   8-bytes  |   8-bytes  |   8-bytes  |
    |____________________________|____________|____________|____________|
    |                                                                   |
    |                  56 == (32 + 8 + 8 + 8)-bytes                     |
    |___________________________________________________________________|
    """
    await asleep()
    return (
        (token_bits(256) << 192)
        ^ (ns_counter() << 128)
        ^ (get_thread_id() << 64)
        ^ get_process_id()
    ).to_bytes(56, BIG)


def canonical_token() -> bytes:
    """
    Combines 256-bits of system entropy, a monotonic nanosecond counter,
    as well as process & thread identifiers into a 56-byte token to
    uniquely identify the current context of the function call. These
    tokens SHOULD NOT be exposed publicly.
     _____________________________________
    |                                     |
    |           Format Diagram:           |
    |_____________________________________|
     ___________________________________________________________________
    |                            |            |            |            |
    |         token_bits         |  time (ns) |  thread ID | process ID |
    |----------------------------|------------|------------|------------|
    |                            |            |            |            |
    |          32-bytes          |   8-bytes  |   8-bytes  |   8-bytes  |
    |____________________________|____________|____________|____________|
    |                                                                   |
    |                  56 == (32 + 8 + 8 + 8)-bytes                     |
    |___________________________________________________________________|
    """
    return (
        (token_bits(256) << 192)
        ^ (ns_counter() << 128)
        ^ (get_thread_id() << 64)
        ^ get_process_id()
    ).to_bytes(56, BIG)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    acanonical_token=acanonical_token,
    achoice=achoice,
    arandom_sleep=arandom_sleep,
    atoken_bits=atoken_bits,
    atoken_bytes=atoken_bytes,
    auniform=auniform,
    canonical_token=canonical_token,
    choice=choice,
    random_sleep=random_sleep,
    token_bits=token_bits,
    token_bytes=token_bytes,
    uniform=uniform,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type responsible for starting an entropy gathering background
thread.
"""

__all__ = ["EntropyDaemon"]


from collections import deque

from aiootp._typing import Typing as t
from aiootp._exceptions import Issue
from aiootp.asynchs import Threads, asleep, s_counter

from .simple import acanonical_token, arandom_sleep


class EntropyDaemon:
    """
    Creates & manages a background thread which asynchronously extracts
    from & seeds new entropy into this module's entropy pools. Mixing
    background threading, asynchrony, system entropy, nanosecond
    timestamps, pseudo random sleeps, & SHA3 hashing with a shared
    global object results in prediction & backtracking resistant, non-
    deterministic effects on entropy generation for the whole package.
    """

    __slots__ = (
        "_cache",
        "_cancel",
        "_currently_mutating_frequency",
        "_daemon",
        "_frequency_mutation_deadline",
        "_gadget",
        "_initial_max_delay",
        "_max_delay",
        "_pool",
    )

    def __init__(
        self,
        *,
        entropy_pool: t.Deque[bytes],
        gadget: t.EntropyHashingType,
        max_delay: t.PositiveRealNumber = 1,
    ) -> None:
        """
        Prepares an instance to safely start a background thread.
        """
        self._pool = entropy_pool
        self._gadget = gadget
        self._cache = deque(maxlen=2)
        self._cancel = False
        self._currently_mutating_frequency = False
        self.set_max_delay(max_delay)

    async def _new_snapshot(self) -> bytes:
        """
        Returns a snapshot of the newest & oldest parts of the instance
        entropy pool prepended with a canonical token.
        """
        return await acanonical_token() + self._pool[0] + self._pool[-1]

    async def _acknowledge_temporary_frequency_deadline(self) -> None:
        """
        Switch off the temporary frequency flags if the duration given
        by the user has elapsed.
        """
        await asleep()
        if not self._currently_mutating_frequency:
            return
        deadline_delta = self._frequency_mutation_deadline - s_counter()
        deadline_reached = deadline_delta <= 0
        delay_skips_deadline = self._max_delay - deadline_delta > 1
        if deadline_reached:
            self._max_delay = self._initial_max_delay
            self._currently_mutating_frequency = False
        elif delay_skips_deadline:
            self._max_delay = deadline_delta  # pragma: no cover

    def set_temporary_max_delay(
        self,
        max_delay: t.PositiveRealNumber = 0.001,
        *,
        duration: t.PositiveRealNumber = 1,
    ) -> t.Self:
        """
        Sets a temporary maximum number of seconds a started entropy
        daemon will pseudo-randomly sleep in-between each iteration. The
        initial instance frequency will be restored after `duration`
        number of seconds.
        """
        if max_delay > 0:
            self._frequency_mutation_deadline = s_counter() + duration
            self._max_delay = max_delay
            self._currently_mutating_frequency = True
            return self
        else:
            raise Issue.value_must("max_delay", "be > 0")

    def set_max_delay(self, max_delay: t.PositiveRealNumber = 1) -> t.Self:
        """
        Sets the maximum number of seconds a started entropy daemon will
        pseudo-randomly sleep in-between each iteration. Setting
        `max_delay` to smaller numbers will cause more cpu power to be
        consumed by the background daemon thread.
        """
        if max_delay > 0:
            self._initial_max_delay = max_delay
            self._max_delay = max_delay
            return self
        else:
            raise Issue.value_must("max_delay", "be > 0")

    async def _araw_loop(self) -> None:
        """
        Takes snapshots of & feeds entropy into the module's entropy
        pools before & after sleeping for a pseudo-random amount of time.
        This is done asynchronously & in a background thread to
        automatically keep the package entropy fresh, & to increase the
        unpredictability & non-determinism of entropy generation.
        """
        while True:
            self._cache.appendleft(await self._new_snapshot())
            entropy = await self._gadget.ahash(*self._cache, size=32)
            self._pool.appendleft(entropy)
            await self._acknowledge_temporary_frequency_deadline()
            await arandom_sleep(self._max_delay)
            if self._cancel:
                return

    def start(self) -> t.Self:
        """
        Runs an entropy updating & gathering thread in the background.

        This supports the package by asynchronously & continuously
        seeding into & extracting new entropy from its entropy pools.
        """
        state = Threads._Manager().list()
        self._daemon = Threads._type(
            target=Threads._arun_func,
            args=(self._araw_loop, state),
        )
        self._daemon.daemon = True
        self._daemon.start()
        return self

    def cancel(self) -> t.Self:
        """
        Cancels the background thread.
        """
        self._cancel = True
        return self


module_api = dict(
    EntropyDaemon=t.add_type(EntropyDaemon),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A configuration type for `GUID`.
"""

__all__ = ["GUIDConfig"]


from hashlib import shake_256

from aiootp._typing import Typing as t
from aiootp._paths import SecurePath, read_salt_file
from aiootp.commons import Config
from aiootp.generics import Domains, hash_bytes


# ensure the device has created a secret static salt for GUIDs
_guid_salt_name = Domains.encode_constant(b"default_guid_salt", size=16)
_guid_salt_path = SecurePath(key=_guid_salt_name, _admin=True)
_guid_salt = read_salt_file(_guid_salt_path)


class GUIDConfig(Config):
    """
    A configuration type for `GUID`.
    """

    __slots__ = (
        "_DEFAULT_KEY",
        "AAD",
        "SIZE",
        "KEY_SIZE",
        "RAW_GUID_CONFIG_ID",
        "PERMUTATION_CONFIG_ID",
        "RawGUID",
        "Permutation",
    )

    slots_types: t.Mapping[str, type] = dict(
        _DEFAULT_KEY=bytes,
        AAD=bytes,
        SIZE=int,
        KEY_SIZE=int,
        RAW_GUID_CONFIG_ID=t.Hashable,
        PERMUTATION_CONFIG_ID=t.Hashable,
        RawGUID=type,
        Permutation=t.PermutationType,
    )

    _DEFAULT_SALT_DOMAIN: t.AnyStr = b"default_guid_salt"

    def _process_raw_guid_config_id(
        self, raw_guid_config_id: t.Optional[t.Hashable]
    ) -> t.Hashable:
        """
        Returns the `RawGUID` config ID if specified. If not, the
        instance's own config ID is assumed to be desired & is returned.
        """
        if raw_guid_config_id is None:
            return self.CONFIG_ID
        else:
            return raw_guid_config_id

    def _process_permutation_config_id(
        self, permutation_config_id: t.Optional[t.Hashable]
    ) -> t.Hashable:
        """
        Returns the permutation config ID if specified. If not, the
        instance's own config ID is assumed to be desired & is returned.
        """
        if permutation_config_id is None:
            return self.CONFIG_ID
        else:
            return permutation_config_id

    def _load_default_key(self) -> bytes:
        """
        Return the static device salt used for GUID creation.
        """
        config_id = repr(self.CONFIG_ID).encode()
        return hash_bytes(
            self._DEFAULT_SALT_DOMAIN,
            config_id,
            self.AAD,
            key=_guid_salt,
            hasher=shake_256,
            size=self.KEY_SIZE,
        )

    def __init__(
        self,
        *,
        config_id: t.Hashable,
        size: int,
        raw_guid_type: type,
        permutation_type: t.PermutationType,
        raw_guid_config_id: t.Optional[t.Hashable] = None,
        permutation_config_id: t.Optional[t.Hashable] = None,
        aad: bytes = b"",
    ) -> None:
        self.CONFIG_ID = config_id
        self.AAD = aad
        self.SIZE = size
        self.KEY_SIZE = permutation_type.key_size(config_id)
        self._DEFAULT_KEY = self._load_default_key()
        self.RAW_GUID_CONFIG_ID = self._process_raw_guid_config_id(
            raw_guid_config_id
        )
        self.PERMUTATION_CONFIG_ID = self._process_permutation_config_id(
            permutation_config_id
        )
        self.RawGUID = raw_guid_type
        self.Permutation = permutation_type


module_api = dict(
    GUIDConfig=t.add_type(GUIDConfig),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Functions for converting random keys or passphrases into a [pseudo-]
random sequences of words.
"""

__all__ = ["akeyed_choices", "keyed_choices", "amnemonic", "mnemonic"]


from aiootp._typing import Typing as t
from aiootp._constants import BIG, INT_BYTES, MIN_KEY_BYTES, WORD_LIST
from aiootp._exceptions import Issue
from aiootp._gentools import abatch, batch
from aiootp.generics import Domains, acanonical_pack, canonical_pack
from aiootp.randoms import acsprng, csprng

from .domain_kdf import DomainKDF
from .passcrypt import Passcrypt


async def akeyed_choices(
    choices: t.Sequence[t.Any],
    selection_size: int,
    *,
    domain: bytes = b"",
    key: bytes,
) -> t.AsyncGenerator[t.Any, None]:
    """
    Makes `selection_size` number of selections from an indexable
    sequence of `choices` using subkeys derived from a provided
    `domain` & `key`. Yields each selection one at a time.
    """
    total_choices = len(choices)
    key = await DomainKDF(
        domain,
        total_choices.to_bytes(INT_BYTES, BIG),
        selection_size.to_bytes(INT_BYTES, BIG),
        key=key,
    ).ashake_256(size=16 * selection_size, aad=Domains.PRNG)
    async for index in abatch(key, size=16):
        yield choices[int.from_bytes(index, BIG) % total_choices]


def keyed_choices(
    choices: t.Sequence[t.Any],
    selection_size: int,
    *,
    domain: bytes = b"",
    key: bytes,
) -> t.Generator[t.Any, None, None]:
    """
    Makes `selection_size` number of selections from an indexable
    sequence of `choices` using subkeys derived from a provided
    `domain` & `key`. Yields each selection one at a time.
    """
    total_choices = len(choices)
    key = DomainKDF(
        domain,
        total_choices.to_bytes(INT_BYTES, BIG),
        selection_size.to_bytes(INT_BYTES, BIG),
        key=key,
    ).shake_256(size=16 * selection_size, aad=Domains.PRNG)
    for index in batch(key, size=16):
        yield choices[int.from_bytes(index, BIG) % total_choices]


async def amnemonic(
    passphrase: t.Optional[bytes] = None,
    size: int = 8,
    *,
    salt: t.Optional[bytes] = None,
    words: t.Optional[t.Sequence[t.Any]] = None,
    **passcrypt_settings: t.Union[int, t.ConfigType],
) -> t.List[bytes]:
    """
    Creates list of `size` number of words for a mnemonic key from a
    user `passphrase` & an optional `salt`. If no `passphrase` is
    supplied, then a random value is used to derive a unique mnemonic.
    The `words` used for the mnemonic can be passed, but by default
    are a word-list of 2048 unique, all lowercase english words.
    """
    domain = Domains.MNEMONIC
    words = words if words else WORD_LIST
    salt = await acanonical_pack(salt if salt else b"", domain)
    if passphrase:
        pcrypt = Passcrypt(**passcrypt_settings, tag_size=MIN_KEY_BYTES)
        key = await pcrypt.anew(passphrase, salt)
    elif not passcrypt_settings:
        key = await acsprng()
    else:
        mistake = "passcrypt_settings", "generating a random mnemonic key"
        raise Issue.unused_parameters(*mistake)
    choices = akeyed_choices(words, size, domain=domain, key=key)
    return [word async for word in choices]


def mnemonic(
    passphrase: t.Optional[bytes] = None,
    size: int = 8,
    *,
    salt: t.Optional[bytes] = None,
    words: t.Optional[t.Sequence[t.Any]] = None,
    **passcrypt_settings: t.Union[int, t.ConfigType],
) -> t.List[bytes]:
    """
    Creates list of `size` number of words for a mnemonic key from a
    user `passphrase` & an optional `salt`. If no `passphrase` is
    supplied, then a random value is used to derive a unique mnemonic.
    The `words` used for the mnemonic can be passed, but by default
    are a word-list of 2048 unique, all lowercase english words.
    """
    domain = Domains.MNEMONIC
    words = words if words else WORD_LIST
    salt = canonical_pack(salt if salt else b"", domain)
    if passphrase:
        pcrypt = Passcrypt(**passcrypt_settings, tag_size=MIN_KEY_BYTES)
        key = pcrypt.new(passphrase, salt)
    elif not passcrypt_settings:
        key = csprng()
    else:
        mistake = "passcrypt_settings", "generating a random mnemonic key"
        raise Issue.unused_parameters(*mistake)
    return [*keyed_choices(words, size, domain=domain, key=key)]


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    akeyed_choices=akeyed_choices,
    amnemonic=amnemonic,
    keyed_choices=keyed_choices,
    mnemonic=mnemonic,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from collections import deque

from conftest import *

from aiootp.asynchs import ConcurrencyGuard


class TestConcurrencyGuard:
    async def test_detects_async_out_of_order_execution(self) -> None:
        problem = (  # fmt: skip
            "Another execution authorization token was allowed to skip "
            "the current's place in line."
        )
        error = ConcurrencyGuard.IncoherentConcurrencyState

        with Ignore(error, if_else=violation(problem)):
            queue = deque()
            async with ConcurrencyGuard(queue):
                queue.appendleft(token_bytes(32))

        with Ignore(error, if_else=violation(problem)):
            queue = deque()
            async with ConcurrencyGuard(queue):
                queue[0] = token_bytes(32)

        with Ignore(IndexError, if_else=violation(problem)):
            queue = deque()
            async with ConcurrencyGuard(queue):
                queue.popleft()

    async def test_detects_sync_out_of_order_execution(self) -> None:
        problem = (  # fmt: skip
            "Another execution authorization token was allowed to skip "
            "the current's place in line."
        )
        error = ConcurrencyGuard.IncoherentConcurrencyState

        with Ignore(error, if_else=violation(problem)):
            queue = deque()
            with ConcurrencyGuard(queue):
                queue.appendleft(token_bytes(32))

        with Ignore(error, if_else=violation(problem)):
            queue = deque()
            with ConcurrencyGuard(queue):
                queue[0] = token_bytes(32)

        with Ignore(IndexError, if_else=violation(problem)):
            queue = deque()
            with ConcurrencyGuard(queue):
                queue.popleft()


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
File IO & bytes encoding utilities.
"""

__all__ = ["ByteIO"]


import aiofiles
from math import ceil
from pathlib import Path
from base64 import urlsafe_b64encode, urlsafe_b64decode

from aiootp._typing import Typing as t
from aiootp._constants import Tables, BIG
from aiootp.commons import FrozenInstance
from aiootp.asynchs import asleep

from .transform import abase_as_int, base_as_int, aint_as_base, int_as_base
from .canon import fullblock_ljust


class ByteIO(FrozenInstance):
    """
    An interface for bytes encodinga & IO operations.
    """

    __slots__ = ()

    @staticmethod
    async def abytes_to_urlsafe(value: bytes) -> bytes:
        """
        Converts a raw bytes `value` to a url safe, base64 encoded
        byte string.
        """
        await asleep()
        return urlsafe_b64encode(value).replace(b"=", b"")

    @staticmethod
    def bytes_to_urlsafe(value: bytes) -> bytes:
        """
        Converts a raw bytes `value` to a url safe, base64 encoded
        byte string.
        """
        urlsafe_value = urlsafe_b64encode(value)
        return urlsafe_value.replace(b"=", b"")

    @staticmethod
    async def aurlsafe_to_bytes(value: t.AnyStr) -> bytes:
        """
        Turns a url safe base64 encoded `value` back into a raw
        decoded byte string.
        """
        await asleep()
        if value.__class__ is str:
            value = value.encode()
        return urlsafe_b64decode(fullblock_ljust(value, 4, pad=b"="))

    @staticmethod
    def urlsafe_to_bytes(value: t.AnyStr) -> bytes:
        """
        Turns a url safe base64 encoded `value` back into a raw
        decoded byte string.
        """
        if value.__class__ is str:
            value = value.encode()
        return urlsafe_b64decode(fullblock_ljust(value, 4, pad=b"="))

    @staticmethod
    async def abytes_to_filename(value: bytes) -> str:
        """
        Returns the received bytes-type `value` in base38 encoding,
        which can be used as a filename to maintain compatibility on a
        very wide array of platforms.
        """
        return await aint_as_base(
            int.from_bytes(value, BIG), base=38, table=Tables.BASE_38
        )

    @staticmethod
    def bytes_to_filename(value: bytes) -> str:
        """
        Returns the received bytes-type `value` in base38 encoding,
        which can be used as a filename to maintain compatibility on a
        very wide array of platforms.
        """
        return int_as_base(
            int.from_bytes(value, BIG), base=38, table=Tables.BASE_38
        )

    @staticmethod
    async def afilename_to_bytes(filename: str) -> bytes:
        """
        Returns the base38 encoded `filename` as raw decoded bytes.
        """
        result = await abase_as_int(filename, base=38, table=Tables.BASE_38)
        byte_count = ceil(result.bit_length() / 8)
        return result.to_bytes(byte_count, BIG)

    @staticmethod
    def filename_to_bytes(filename: str) -> bytes:
        """
        Returns the base38 encoded `filename` as raw decoded bytes.
        """
        result = base_as_int(filename, base=38, table=Tables.BASE_38)
        byte_count = ceil(result.bit_length() / 8)
        return result.to_bytes(byte_count, BIG)

    @staticmethod
    async def aread(path: t.PathStr) -> bytes:
        """
        Reads the bytes data from the file at `path`.
        """
        async with aiofiles.open(path, "rb") as f:
            return await f.read()

    @staticmethod
    def read(path: t.PathStr) -> bytes:
        """
        Reads the bytes data from the file at `path`.
        """
        with Path(path).open("rb") as f:
            return f.read()

    @staticmethod
    async def awrite(path: t.PathStr, data: bytes) -> None:
        """
        Writes bytes `data` to the file at `path`.
        """
        async with aiofiles.open(path, "wb+") as f:
            await f.write(data)

    @staticmethod
    def write(path: t.PathStr, data: bytes) -> None:
        """
        Writes bytes `data` to the file at `path`.
        """
        with Path(path).open("wb+") as f:
            f.write(data)

    @staticmethod
    async def aappend(path: t.PathStr, data: bytes) -> None:
        """
        Appends bytes `data` to the file at `path`.
        """
        async with aiofiles.open(path, "ab+") as f:
            await f.write(data)

    @staticmethod
    def append(path: t.PathStr, data: bytes) -> None:
        """
        Appends bytes `data` to the file at `path`.
        """
        with Path(path).open("ab+") as f:
            f.write(data)


module_api = dict(
    ByteIO=t.add_type(ByteIO),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Streaming manager classes for the package's ciphers.
"""

__all__ = ["AsyncDecipherStream", "DecipherStream"]


import io
from collections import deque
from hmac import compare_digest
from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._constants import DEFAULT_AAD, DEFAULT_TTL
from aiootp._exceptions import Issue, CipherStreamIsClosed
from aiootp._gentools import apopleft, popleft, abatch, batch
from aiootp.asynchs import AsyncInit, ConcurrencyGuard, asleep

from .cipher_stream_properties import AuthFail, CipherStreamProperties


class AsyncDecipherStream(CipherStreamProperties, metaclass=AsyncInit):
    """
    A high-level public interface to the package's online, salt reuse /
    misuse resistant, tweakable, AEAD ciphers.

    Validation using `StreamHMAC` `block_id`'s provides robustness to
    communication channels. It does so by detecting forgeries & out-of-
    order messages without changing its internal state. If a tag fails
    to validate, the cipher stream can continue functioning once the
    authentic messages, in their correct order, begin coming in.
    --------
    HOWEVER: The more blocks that are allowed to fail without aborting,
    -------- the more chances an attacker has to spoof the auth tag.

     _____________________________________
    |                                     |
    |      Usage Example: Decryption      |
    |_____________________________________|


    stream = await AsyncDecipherStream(
        key, salt=session.salt, aad=session.transcript, iv=session.iv
    )
    for ciphertext in session.download.buffer(4 * stream.PACKETSIZE):
        await stream.abuffer(ciphertext)  # <--- raises InvalidBlockID if
        async for plaintext in stream:    # auth failure in last 4 packets
            yield plaintext

    async for plaintext in stream.afinalize():
        yield plaintext

    # Verify the stream termination / authentication tag. <--------
    await stream.shmac.atest_shmac(session.shmac)
    """

    __slots__ = (
        "_buffer",
        "_bytes_to_trim",
        "_config",
        "_digesting_now",
        "_finalizing_now",
        "_is_streaming",
        "_key_bundle",
        "_padding",
        "_result_queue",
        "_stream",
        "_ttl",
        "shmac",
    )

    _MAX_SIMULTANEOUS_BUFFERS: int = 1024

    async def __init__(
        self,
        cipher: t.CipherInterfaceType,
        *,
        salt: bytes,
        aad: bytes = DEFAULT_AAD,
        iv: bytes,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> None:
        """
        Derives decryption keys & initializes a mutable buffer to
        automatically decrypt & return plaintext with padding removed.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `iv`: An ephemeral, uniform, random value that's generated by
                the encryption algorithm.

        `ttl`: An amount of seconds that dictate the allowable age of
                a ciphertext stream, but ONLY checks the time for expiry
                at the very start of a stream. Has no effect on how long
                a stream can continue to live for.
        """
        self._config = cipher._config
        self._padding = cipher._padding
        self._ttl = ttl
        self._digesting_now = deque(maxlen=self._MAX_SIMULTANEOUS_BUFFERS)
        self._finalizing_now = deque()  # don't let maxlen remove entries
        self._is_streaming = False
        self._result_queue = deque()
        self._buffer = buffer = deque()
        self._bytes_to_trim = self._config.INNER_HEADER_BYTES
        self._key_bundle = key_bundle = await cipher._KeyAADBundle(
            kdfs=cipher._kdfs,
            salt=salt,
            aad=aad,
            iv=iv,
        ).async_mode()
        self.shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
        self._stream = cipher._Junction.abytes_decipher(
            apopleft(buffer), shmac=self.shmac
        )

    @property
    def _iter_shortcuts(
        self,
    ) -> t.Tuple[t.Deque[bytes], t.Callable[[], bytes]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self._result_queue, self._result_queue.popleft

    @property
    def _digest_data_shortcuts(
        self,
    ) -> t.Tuple[t.Callable[[None], bytes], t.Callable[[bytes], None]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self._stream.asend, self._result_queue.append

    @property
    def _buffer_shortcuts(
        self,
    ) -> t.Tuple[t.Callable[..., None], t.Callable[[bytes], None]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self.shmac.atest_next_block_id, self._buffer.append

    async def _atest_timestamp(
        self, queue: t.Callable[[bytes], None]
    ) -> None:
        """
        Raises `TimestampExpired` if the timestamp prepended to the
        plaintext is older than the time-to-live specified by the
        instance.
        """
        try:
            self._is_streaming = True
            timestamp = queue[0][self._config.TIMESTAMP_SLICE]
            await self._config.clock.atest_timestamp(timestamp, self._ttl)
        except self.TimestampExpired as error:
            self._is_streaming = False
            raise error

    async def _aremove_inner_header(
        self, queue: t.Callable[[bytes], None]
    ) -> None:
        """
        Strips the inner header from the buffered plaintext in the queue
        in the cases where the inner header spans multiple blocks.
        """
        await asleep()
        inner_header = queue[0][: self._bytes_to_trim]
        block = queue[0][self._bytes_to_trim :]
        self._bytes_to_trim -= len(inner_header)
        if block:
            queue[0] = block
        else:
            queue.popleft()

    async def __aiter__(self) -> t.AsyncGenerator[bytes, None]:
        """
        Allows the object to be entered in for-loops an unlimited amount
        of times in the process of gathering data to buffer for the
        stream. This leads to a very pythonic API.
        """
        MIN_STREAM_QUEUE = self._config.MIN_STREAM_QUEUE
        result_queue, pop_result = self._iter_shortcuts
        if not self._is_streaming:
            await self._atest_timestamp(result_queue)
        while self._bytes_to_trim and result_queue:
            await self._aremove_inner_header(result_queue)
        while len(result_queue) > MIN_STREAM_QUEUE:
            yield pop_result()
            await asleep()

    async def afinalize(self) -> t.AsyncGenerator[bytes, None]:
        """
        Instructs the instance to finish receiving data into the buffer
        & to flush all results out to the user with its plaintext
        padding removed.

         _____________________________________
        |                                     |
        |      Usage Example: Decryption      |
        |_____________________________________|


        for ciphertext in session.download.buffer(4 * stream.PACKETSIZE):
            await stream.abuffer(ciphertext)  # <--- raises InvalidBlockID if
            async for plaintext in stream:    # auth failure in last 4 packets
                yield plaintext

        async for plaintext in stream.afinalize():
            yield plaintext
        """
        self._finalizing_now.append(token := token_bytes(32))
        if not compare_digest(token, self._finalizing_now[0]):
            raise ConcurrencyGuard.IncoherentConcurrencyState

        async with ConcurrencyGuard(self._digesting_now, token=token):
            await self.shmac.afinalize()
            async for result in self:
                yield result
            queue = self._result_queue
            footer_index = await self._padding.adepadding_end_index(
                queue[-1]
            )
            async for block in abatch(
                b"".join(queue)[:footer_index], size=self._config.BLOCKSIZE
            ):
                yield block

    async def _adigest_data(
        self,
        data: t.Callable[[int], bytes],
        atest_block_id: t.Callable[..., None],
        append: t.Callable[[bytes], None],
    ) -> None:
        """
        Prepares the input ciphertext `data` for decryption by dividing
        it into blocksize chunks & validating each packet's block ID.
        """
        config = self._config
        BLOCKSIZE, BLOCK_ID_BYTES = config.BLOCKSIZE, config.BLOCK_ID_BYTES
        cipher, queue_result = self._digest_data_shortcuts
        while True:
            block_id = data(BLOCK_ID_BYTES)
            if not block_id:
                break
            block = data(BLOCKSIZE)
            try:
                await atest_block_id(block_id, block)
            except self.InvalidBlockID as auth_fail:
                # Package the current state of buffering for the caller
                # to handle authentication failures & negotiation of
                # data retransmission if desired.
                auth_fail.failure_state = AuthFail(block_id, block, data)
                raise auth_fail
            append(block)
            queue_result(await cipher(None))

    async def abuffer(self, data: bytes) -> t.Self:
        """
        Prepares the input ciphertext `data` for decryption by
        dividing it into blocksize chunks & validating each packet's
        block ID.

         _____________________________________
        |                                     |
        |      Usage Example: Decryption      |
        |_____________________________________|


        for ciphertext in session.download.buffer(4 * stream.PACKETSIZE):
            await stream.abuffer(ciphertext)  # <--- raises InvalidBlockID if
            async for plaintext in stream:    # auth failure in last 4 packets
                yield plaintext

        async for plaintext in stream.afinalize():
            yield plaintext
        """
        if not data or len(data) % self.PACKETSIZE:
            raise Issue.invalid_length("data", len(data))

        async with ConcurrencyGuard(self._digesting_now):
            if await self._aconstant_time_final_context_is_done():
                raise CipherStreamIsClosed
            data = io.BytesIO(data).read
            atest_block_id, append = self._buffer_shortcuts
            await self._adigest_data(data, atest_block_id, append)
        return self


class DecipherStream(CipherStreamProperties):
    """
    A high-level public interface to the package's online, salt reuse /
    misuse resistant, tweakable, AEAD ciphers.

    Validation using `StreamHMAC` `block_id`'s provides robustness to
    communication channels. It does so by detecting forgeries & out-of-
    order messages without changing its internal state. If a tag fails
    to validate, the cipher stream can continue functioning once the
    authentic messages, in their correct order, begin coming in.
    --------
    HOWEVER: The more blocks that are allowed to fail without aborting,
    -------- the more chances an attacker has to spoof the auth tag.

     _____________________________________
    |                                     |
    |      Usage Example: Decryption      |
    |_____________________________________|


    stream = DecipherStream(
        key, salt=session.salt, aad=session.transcript, iv=session.iv
    )
    for ciphertext in session.download.buffer(4 * stream.PACKETSIZE):
        stream.buffer(ciphertext)   # <--- raises InvalidBlockID if
        for plaintext in stream:    # auth failure in last 4 packets
            yield plaintext

    for plaintext in stream.finalize():
        yield plaintext

    # Verify the stream termination / authentication tag. <--------
    stream.shmac.test_shmac(session.shmac)
    """

    __slots__ = (
        "_buffer",
        "_config",
        "_bytes_to_trim",
        "_digesting_now",
        "_finalizing_now",
        "_is_streaming",
        "_key_bundle",
        "_padding",
        "_result_queue",
        "_stream",
        "_ttl",
        "shmac",
    )

    _MAX_SIMULTANEOUS_BUFFERS: int = 1024

    def __init__(
        self,
        cipher: t.CipherInterfaceType,
        *,
        salt: bytes,
        aad: bytes = DEFAULT_AAD,
        iv: bytes,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> None:
        """
        Derives decryption keys & initializes a mutable buffer to
        automatically decrypt & return plaintext with padding removed.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `iv`: An ephemeral, uniform, random value that's generated by
                the encryption algorithm.

        `ttl`: An amount of seconds that dictate the allowable age of
                a ciphertext stream, but ONLY checks the time for expiry
                at the very start of a stream. Has no effect on how long
                a stream can continue to live for.
        """
        self._config = cipher._config
        self._padding = cipher._padding
        self._ttl = ttl
        self._digesting_now = deque(maxlen=self._MAX_SIMULTANEOUS_BUFFERS)
        self._finalizing_now = deque()  # don't let maxlen remove entries
        self._is_streaming = False
        self._result_queue = deque()
        self._buffer = buffer = deque()
        self._bytes_to_trim = self._config.INNER_HEADER_BYTES
        self._key_bundle = key_bundle = cipher._KeyAADBundle(
            kdfs=cipher._kdfs,
            salt=salt,
            aad=aad,
            iv=iv,
        ).sync_mode()
        self.shmac = cipher._StreamHMAC(key_bundle)._for_decryption()
        self._stream = cipher._Junction.bytes_decipher(
            popleft(buffer), shmac=self.shmac
        )

    @property
    def _iter_shortcuts(
        self,
    ) -> t.Tuple[t.Deque[bytes], t.Callable[[], bytes]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self._result_queue, self._result_queue.popleft

    @property
    def _digest_data_shortcuts(
        self,
    ) -> t.Tuple[t.Callable[[None], bytes], t.Callable[[bytes], None]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self._stream.send, self._result_queue.append

    @property
    def _buffer_shortcuts(
        self,
    ) -> t.Tuple[t.Callable[..., None], t.Callable[[bytes], None]]:
        """
        Returns method pointers so calls in tight loops during
        processing don't have to continually getattr on instance
        objects.
        """
        return self.shmac.test_next_block_id, self._buffer.append

    def _test_timestamp(self, queue: t.Callable[[bytes], None]) -> None:
        """
        Raises `TimestampExpired` if the timestamp prepended to the
        plaintext is older than the time-to-live specified by the
        instance.
        """
        try:
            self._is_streaming = True
            timestamp = queue[0][self._config.TIMESTAMP_SLICE]
            self._config.clock.test_timestamp(timestamp, self._ttl)
        except self.TimestampExpired as error:
            self._is_streaming = False
            raise error

    def _remove_inner_header(
        self, queue: t.Callable[[bytes], None]
    ) -> None:
        """
        Strips the inner header from the buffered plaintext in the queue
        in the cases where the inner header spans multiple blocks.
        """
        inner_header = queue[0][: self._bytes_to_trim]
        block = queue[0][self._bytes_to_trim :]
        self._bytes_to_trim -= len(inner_header)
        if block:
            queue[0] = block
        else:
            queue.popleft()

    def __iter__(self) -> t.Generator[bytes, None, None]:
        """
        Allows the object to be entered in for-loops an unlimited amount
        of times in the process of gathering data to buffer for the
        stream. This leads to a very pythonic API.
        """
        MIN_STREAM_QUEUE = self._config.MIN_STREAM_QUEUE
        result_queue, pop_result = self._iter_shortcuts
        if not self._is_streaming:
            self._test_timestamp(result_queue)
        while self._bytes_to_trim and result_queue:
            self._remove_inner_header(result_queue)
        while len(result_queue) > MIN_STREAM_QUEUE:
            yield pop_result()

    def finalize(self) -> t.Generator[bytes, None, None]:
        """
        Instructs the instance to finish receiving data into the buffer
        & to flush all results out to the user with its plaintext
        padding removed.

         _____________________________________
        |                                     |
        |      Usage Example: Decryption      |
        |_____________________________________|

        for ciphertext in source.download.buffer(4 * stream.PACKETSIZE):
            stream.buffer(ciphertext)   # <--- raises InvalidBlockID if
            for plaintext in stream:    # auth failure in last 4 packets
                yield plaintext
        for plaintext in stream.finalize():
            yield plaintext
        """
        self._finalizing_now.append(token := token_bytes(32))
        if not compare_digest(token, self._finalizing_now[0]):
            raise ConcurrencyGuard.IncoherentConcurrencyState

        with ConcurrencyGuard(
            self._digesting_now, probe_delay=0.0001, token=token
        ):
            self.shmac.finalize()
            yield from self
            queue = self._result_queue
            footer_index = self._padding.depadding_end_index(queue[-1])
            yield from batch(
                b"".join(queue)[:footer_index], size=self._config.BLOCKSIZE
            )

    def _digest_data(
        self,
        data: t.Callable[[int], bytes],
        test_block_id: t.Callable[..., None],
        append: t.Callable[[bytes], None],
    ) -> None:
        """
        Prepares the input ciphertext `data` for decryption by dividing
        it into blocksize chunks & validating each packet's block ID.
        """
        config = self._config
        BLOCKSIZE, BLOCK_ID_BYTES = config.BLOCKSIZE, config.BLOCK_ID_BYTES
        cipher, queue_result = self._digest_data_shortcuts
        while True:
            block_id = data(BLOCK_ID_BYTES)
            if not block_id:
                break
            block = data(BLOCKSIZE)
            try:
                test_block_id(block_id, block)
            except self.InvalidBlockID as auth_fail:
                # Package the current state of buffering for the caller
                # to handle authentication failures & negotiation of
                # data retransmission if desired.
                auth_fail.failure_state = AuthFail(block_id, block, data)
                raise auth_fail
            append(block)
            queue_result(cipher(None))

    def buffer(self, data: bytes) -> t.Self:
        """
        Prepares the input ciphertext `data` for decryption by
        dividing it into blocksize chunks & validating each packet's
        block ID.

         _____________________________________
        |                                     |
        |      Usage Example: Decryption      |
        |_____________________________________|

        for ciphertext in source.download.buffer(4 * stream.PACKETSIZE):
            stream.buffer(ciphertext)   # <--- raises InvalidBlockID if
            for plaintext in stream:    # auth failure in last 4 packets
                yield plaintext
        for plaintext in stream.finalize():
            yield plaintext
        """
        if not data or len(data) % self.PACKETSIZE:
            raise Issue.invalid_length("data", len(data))

        with ConcurrencyGuard(self._digesting_now, probe_delay=0.0001):
            if self._constant_time_final_context_is_done():
                raise CipherStreamIsClosed
            data = io.BytesIO(data).read
            atest_block_id, append = self._buffer_shortcuts
            self._digest_data(data, atest_block_id, append)
        return self


module_api = dict(
    AsyncDecipherStream=t.add_type(AsyncDecipherStream),
    DecipherStream=t.add_type(DecipherStream),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Interfaces for creating, deriving, & using symmetric & asymmetric
cryptographic keys.
"""

__all__ = [
    "DomainKDF",
    "Ed25519",
    "PackageSigner",
    "PackageVerifier",
    "Passcrypt",
    "X25519",
    "amnemonic",
    "mnemonic",
]


from .domain_kdf import *
from .passcrypt import *
from .mnemonics import *
from .curve_25519 import *
from .package_signer import *
from .package_verifier import *


subpackages = dict(curve_25519=curve_25519, passcrypt=passcrypt)


modules = dict(
    domain_kdf=domain_kdf,
    mnemonics=mnemonics,
    package_signer=package_signer,
    package_verifier=package_verifier,
)


module_api = dict(
    DomainKDF=DomainKDF,
    Ed25519=Ed25519,
    PackageSigner=PackageSigner,
    PackageVerifier=PackageVerifier,
    Passcrypt=Passcrypt,
    X25519=X25519,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    amnemonic=amnemonic,
    mnemonic=mnemonic,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A chain of `AffinePermutation` objects with distinct keys & which
have their inputs & outputs XOR'd with key material.
"""

__all__ = ["AffineXORChain"]


import io

from aiootp._typing import Typing as t
from aiootp._constants import BIG
from aiootp._exceptions import Issue
from aiootp.asynchs import asleep
from aiootp.commons import ConfigMap, FrozenInstance

from .affine_permutation import AffinePermutation
from .affine_xor_chain_config import AffineXORChainConfig


class AffineXORChain(FrozenInstance):
    """
    Creates instances which perform invertible, bijective, keyed affine
    permutations on integers within a given byte-domain. XORing keys
    with the inputs & outputs of chained affine permutations helps to
    break the algebraic structure of the permutation.

    The `(a)step` methods adjust the permutation such that repeat inputs
    within the domain will never result in a repeated output if no input
    is repeated prior to a call of the methods & the number of calls to
    the methods doesn't exceed the size of the domain.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp._permutations.affine_xor_chain import AffineXORChain

    domain_size = 1
    key = token_bytes(AffineXORChain.key_size(domain_size))
    aff = AffineXORChain(key=key, config_id=domain_size)

    latin_square = []
    domain = tuple(range(256 ** domain_size))
    for _ in domain:
        row = tuple(aff.permute(element) for element in domain)
        latin_square.append(row)
        aff.step()

    assert len(set(latin_square)) == len(domain)
    assert all(len(set(row)) == len(domain) for row in latin_square)
    columns = [[row[column] for row in latin_square] for column in domain]
    assert all(len(set(column)) == len(domain) for column in columns)
    """

    __slots__ = (
        "_aff_in",
        "_aff_mid",
        "_aff_out",
        "_in_key",
        "_in_mid_key",
        "_out_mid_key",
        "_out_key",
        "_increment",
        "config",
    )

    _configs = ConfigMap(
        mapping={
            config_id: AffineXORChainConfig(
                config_id=config_id,
                size=config_id,
                permutation_type=AffinePermutation,
                permutation_config_id=config_id,
            )
            for config_id in [*range(1, 33), 64, 128, 192, 256]
        },
        config_type=AffineXORChainConfig,
    )

    @classmethod
    def key_size(cls, config_id: t.Hashable) -> int:
        """
        Returns the integer size of the initialization key an instance
        uses which is defined by the configuration indexed by the
        `config_id`.
        """
        return cls._configs[config_id].KEY_SIZE

    def _ingest_key(self, key: bytes) -> t.Callable[[int], bytes]:
        """
        Returns a byte reader that's used to segment the input `key` if
        it's the correct size, otherwise raises `ValueError`.
        """
        key_size = self.key_size(self.config.CONFIG_ID)
        if len(key) == key_size:
            return io.BytesIO(key).read
        else:
            cls = self.__class__.__qualname__
            raise Issue.invalid_length(f"{cls} key", key_size)

    def _process_in_key(
        self, key_reader: t.Callable[[int], bytes], size: int
    ) -> int:
        """
        Returns an integer version of a segment of the instance's key
        if the segment retrieved is the correct size, otherwise raises
        `ValueError`.
        """
        key = key_reader(size)
        if len(key) == size:
            return int.from_bytes(key, BIG)
        else:
            raise Issue.invalid_length("in xor key", size)

    def _process_mid_key(
        self, key_reader: t.Callable[[int], bytes], size: int
    ) -> t.Tuple[int, int]:
        """
        Returns two integer versions with optimal hamming distance of a
        segment of the instance's key. If the segment retrieved isn't
        the correct size, `ValueError` is raised. It's ensured that at
        least one of these keys is non-zero, with a sufficient amount of
        bits set, so they can't incidentally fail at removing algebraic
        structure from the permutation.
        """
        c = self.config
        key = key_reader(size)
        if len(key) == size:
            preprocessed_key = int.from_bytes(key, BIG)
            return preprocessed_key ^ c.IPAD, preprocessed_key ^ c.OPAD
        else:
            raise Issue.invalid_length("mid xor key", size)

    def _process_out_key(
        self, key_reader: t.Callable[[int], bytes], size: int
    ) -> int:
        """
        Returns an integer version of a segment of the instance's key
        if the segment retrieved is the correct size, otherwise raises
        `ValueError`.
        """
        key = key_reader(size)
        if len(key) == size:
            return int.from_bytes(key, BIG)
        else:
            raise Issue.invalid_length("out xor key", size)

    def _initialize_keys(
        self, key_reader: t.Callable[[int], bytes], size: int
    ) -> t.Tuple[bytes, bytes, bytes, bytes]:
        """
        Returns the instance's integer keys that are used to XOR the
        inputs & outputs between the chained calls to its component
        affine permutations.
        """
        return (
            self._process_in_key(key_reader, size),
            *self._process_mid_key(key_reader, size),
            self._process_out_key(key_reader, size),
        )

    def _initialize_permutations(
        self, key_reader: t.Callable[[int], bytes], size: int
    ) -> t.Tuple[t.PermutationType]:
        """
        Uses the permutation configuration ID saved in the instance's
        config to initialize & return the affine permutations which will
        be chained together.
        """
        config_id = self.config.PERMUTATION_CONFIG_ID
        Permutation = self.config.Permutation
        return (
            Permutation(key=key_reader(size), config_id=config_id),
            Permutation(key=key_reader(size), config_id=config_id),
            Permutation(key=key_reader(), config_id=config_id),
        )

    def __init__(self, *, key: bytes, config_id: t.Hashable = 16) -> None:
        """
        Uses a bytes-type `key`, which is the correct size as defined
        by the configuration object that's indexed by the `config_id`,
        to set up the keys & permutations to be chained together.
        """
        self.config = c = self._configs[config_id]
        self._increment = c.XPAD
        key_reader = self._ingest_key(key)
        self._in_key, self._in_mid_key, self._out_mid_key, self._out_key = (
            self._initialize_keys(key_reader, c.XOR_KEY_SIZE)
        )
        self._aff_in, self._aff_mid, self._aff_out = (
            self._initialize_permutations(key_reader, c.ADD_KEY_SIZE)
        )

    async def aupdate_increment(self, value: int, /) -> None:
        """
        Augments the instance's step function increment with `value`.
        """
        await asleep()
        self.update_increment(value)

    def update_increment(self, value: int, /) -> None:
        """
        Augments the instance's step function increment with `value`.
        """
        if self.config.DOMAIN > value > 0:
            new_increment = (self._increment ^ value) | 1  # Ensure non-even
            object.__setattr__(self, "_increment", new_increment)
        else:
            raise Issue.value_must("adjustment", "be < DOMAIN & > 0")

    async def astep(self) -> None:
        """
        Tweaks the permutation in a way which can fill a complete latin
        square when the number of distinct consecutive inputs between
        each step, & the number of steps, is equal to the size of the
        domain.
        """
        await asleep()
        self.step()

    def step(self) -> None:
        """
        Tweaks the permutation in a way which can fill a complete latin
        square when the number of distinct consecutive inputs between
        each step, & the number of steps, is equal to the size of the
        domain.
        """
        new_in_key = (self._in_key + self._increment) % self.config.DOMAIN
        object.__setattr__(self, "_in_key", new_in_key)

    async def auncapped_permute(self, value: int, /) -> int:
        """
        Chains keyed bijective affine permutations together while
        XORing each result between the components with keys designed to
        break up the algebraic structure of the permutation.
        """
        return await self._aff_out.apermute(
            await self._aff_mid.apermute(
                await self._aff_in.apermute(value) ^ self._in_mid_key
            )
            ^ self._out_mid_key
        )

    def uncapped_permute(self, value: int, /) -> int:
        """
        Chains keyed bijective affine permutations together while
        XORing each result between the components with keys designed to
        break up the algebraic structure of the permutation.
        """
        return self._aff_out.permute(
            self._aff_mid.permute(
                self._aff_in.permute(value) ^ self._in_mid_key
            )
            ^ self._out_mid_key
        )

    async def apermute(self, value: int, /) -> int:
        """
        Chains keyed bijective affine permutations together while
        additionally XORing each of their inputs & outputs with keys
        designed to break up the algebraic structure of the permutation.
        """
        return (
            await self._aff_out.apermute(
                await self._aff_mid.apermute(
                    await self._aff_in.apermute(value ^ self._in_key)
                    ^ self._in_mid_key
                )
                ^ self._out_mid_key
            )
            ^ self._out_key
        )

    def permute(self, value: int, /) -> int:
        """
        Chains keyed bijective affine permutations together while
        additionally XORing each of their inputs & outputs with keys
        designed to break up the algebraic structure of the permutation.
        """
        return (
            self._aff_out.permute(
                self._aff_mid.permute(
                    self._aff_in.permute(value ^ self._in_key)
                    ^ self._in_mid_key
                )
                ^ self._out_mid_key
            )
            ^ self._out_key
        )

    async def auncapped_invert(self, value: int, /) -> int:
        """
        Inverts the modified chained keyed bijective affine permutation
        done without XORs to the in & out values of the chain.
        """
        return await self._aff_in.ainvert(
            await self._aff_mid.ainvert(
                await self._aff_out.ainvert(value) ^ self._out_mid_key
            )
            ^ self._in_mid_key
        )

    def uncapped_invert(self, value: int, /) -> int:
        """
        Inverts the modified chained keyed bijective affine permutation
        done without XORs to the in & out values of the chain.
        """
        return self._aff_in.invert(
            self._aff_mid.invert(
                self._aff_out.invert(value) ^ self._out_mid_key
            )
            ^ self._in_mid_key
        )

    async def ainvert(self, value: int, /) -> int:
        """
        Inverts the modified chained keyed bijective affine permutation.
        """
        return (
            await self._aff_in.ainvert(
                await self._aff_mid.ainvert(
                    await self._aff_out.ainvert(value ^ self._out_key)
                    ^ self._out_mid_key
                )
                ^ self._in_mid_key
            )
            ^ self._in_key
        )

    def invert(self, value: int, /) -> int:
        """
        Inverts the modified chained keyed bijective affine permutation.
        """
        return (
            self._aff_in.invert(
                self._aff_mid.invert(
                    self._aff_out.invert(value ^ self._out_key)
                    ^ self._out_mid_key
                )
                ^ self._in_mid_key
            )
            ^ self._in_key
        )


module_api = dict(
    AffineXORChain=t.add_type(AffineXORChain),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for the passcrypt memory-hard passphrase hashing
function.
"""

__all__ = ["Passcrypt", "PasscryptSettings"]


from hashlib import shake_128

from aiootp._typing import Typing as t
from aiootp._constants import NS_TO_S_RATIO, DEFAULT_AAD, DEFAULT_TTL
from aiootp._exceptions import Issue, PasscryptIssue, TimestampExpired
from aiootp._exceptions import InvalidPassphrase, ImproperPassphrase
from aiootp.asynchs import Processes
from aiootp.commons import OpenFrozenSlots, FrozenInstance
from aiootp.generics import ahash_bytes, hash_bytes, bytes_are_equal
from aiootp.randoms import acsprng, csprng

from .config import passcrypt_spec
from .hash_format import PasscryptHash
from .sessions_manager import PasscryptProcesses


class PasscryptSettings(OpenFrozenSlots):
    """
    Efficiently stores the resource values located in the header of a
    `Passcrypt` hash.
    """

    __slots__ = ("mb", "cpu", "cores", "tag_size", "salt_size")

    _MAPPED_ATTRIBUTES: t.Tuple[str] = __slots__

    def __init__(
        self,
        *,
        mb: int,
        cpu: int,
        cores: int,
        tag_size: int,
        salt_size: int,
        config: t.ConfigType = passcrypt_spec,
    ) -> None:
        self.mb = mb
        self.cpu = cpu
        self.cores = cores
        self.tag_size = tag_size
        self.salt_size = salt_size
        config.validate_settings(**self)


class Passcrypt(FrozenInstance):
    """
    This class is used to implement an Argon2i-like passphrase-based
    key derivation function that's designed to be resistant to cache-
    timing side-channel attacks & time-memory trade-offs.

    It uses a passphrase-keyed scanning function which sequentially
    passes over unique memory caches requiring a tunable amount of
    difficulty, which is designed here to be very intuitive.

    This scheme is secret independent with regard to how it chooses to
    pass over memory. Through proofs of work & memory, it ensures an
    attacker attempting to crack a passphrase hash cannot complete the
    algorithm substantially faster by storing more memory than what's
    already necessary, or with substantially less memory, by dropping
    cache entries, without drastically increasing the computational
    cost.

    The algorithm initializes all of the columns for the cache using a
    single `shake_128` object after being fed the passphrase, salt, aad
    & all of the parameters. The number of columns is computed
    dynamically to reach the specified memory cost considering that each
    row will hold 2 * max([1, cpu // 2]) digests of 168-bytes. This
    allows the cache to be efficiently allocated up front, benefiting
    further by not needing to resize the memory cache throughout the
    running of the algorithm.

    The sequential passes involve a current row index, the index of the
    row which is the reflection of the first across the cache, & a
    current offset into a row which are multiples of 336 (two digests).
    The index & reflection pointers interleave each other, hashing rows
    with the same object as they scan, & overwriting the 168-byte digest
    / piece of cache at the offset they're pointing to after each hash.

     _____________________________________
    |                                     |
    |    Algorithm Diagram: Side-View     |
    |_____________________________________|

           ___________________ # of rows ___________________
          |                                                 |
          |              initial memory cache               |
          |  row  # of columns == 2 * max([1, cpu // 2])    |
          |   |   # of rows == ⌈1024*1024*mb/168*columns⌉   |
          v   v                                             v
    column|---'-----------------------------------------'---| the initial cache
    column|---'-----------------------------------------'---| of size ~`mb` is
    column|---'-----------------------------------------'---| built very quickly
    column|---'-----------------------------------------'---| using SHAKE-128.
    column|---'-----------------------------------------'---| each (row, column)
    column|---'-----------------------------------------'---| coordinate holds
    column|---'-----------------------------------------'---| one element of
    column|---'-----------------------------------------'---| 168-bytes.
                                                        ^
                                                        |
                           reflection                  row
                          <-   |
          |--------------------'-------'--------------------| each row is
          |--------------------'-------'--------------------| hashed then has
          |--------------------'-------'--------------------| a new 168-byte
          |--------------------'-------'--------------------| digest overwrite
          |--------------------'-------'--------------------| the current pointer
          |--------------------'-------'--------------------| in an alternating
          |--------------------Xxxxxxxx'xxxxxxxxxxxxxxxxxxxx| sequence, first at
          |oooooooooooooooooooo'oooooooO--------------------| the index, then at
                                       |   ->                 its reflection.
                                     index


          |--'-------------------------------------------'--| this continues
          |--'-------------------------------------------'--| until the entire
          |--'-------------------------------------------Xxx| cache has been
          |ooO-------------------------------------------'--| overwritten.
          |xx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'xx| a single `shake_128`
          |oo'ooooooooooooooooooooooooooooooooooooooooooo'oo| object (H) is used
          |xx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'xx| to do all of the
          |oo'ooooooooooooooooooooooooooooooooooooooooooo'oo| hashing.
             |   ->                                 <-   |
           index                                     reflection


          |xxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| finally, the whole
          |ooooooooooo'ooooooooooooooooooooooooooooooooooooo| cache is quickly
          |xxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| hashed `cpu` + 2
          |ooooooooooo'ooooooooooooooooooooooooooooooooooooo| number of times.
          |Fxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| after each pass an
          |foooooooooo'ooooooooooooooooooooooooooooooooooooo| 84-byte digest (F)
          |fxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| is inserted into the
          |foooooooooo'ooooooooooooooooooooooooooooooooooooo| cache, ruling out
                      |   ->                                  hashing state cycles.
                      | hash cpu + 2 # of times               Then a `tag_size`-
                      v                                       byte tag is output.
           H.update(cache)

          tag = H.digest(tag_size)

         _____________________________________
        |                                     |
        |   Format Diagram: Passcrypt Hash    |
        |_____________________________________|
         ______________________________________________________________
        |                                    |                         |
        |               Header               |          Body           |
        |-------|-------|------|------|------|--------|----------------|
        | time  |  mb   | cpu  | cores| Slen |  salt  |      tag       |
        |  8-   |  3-   |  1-  |  1-  |  1-  |  Slen- |     >=16-      |
        | bytes | bytes | byte | byte | byte |  bytes |     bytes      |
        |_______|_______|______|______|______|________|________________|
        |                                                              |
        |                          >=34-bytes                          |
        |______________________________________________________________|

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import Passcrypt, Domains
        from getpass import getpass


        Passcrypt.PEPPER = bytes.fromhex(getpass("hexidecimal pepper: "))
        pcrypt = Passcrypt(mb=128)

        # registration ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        pw_hash = pcrypt.hash_passphrase(pw, aad=un, mb=128)


        # a login attempt ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        try:
            Passcrypt.verify(pw_hash, pw, aad=un, ttl=24 * 3600)
        except Passcrypt.InvalidPassphrase as auth_fail:
            app.post_mortem(error=auth_fail)
        except Passcrypt.TimestampExpired as expired_hash:
            # 24-hour registration expired
            app.post_mortem(error=expired_hash)
    """

    __slots__ = ("_settings", "_config")

    _PasscryptProcesses: type = PasscryptProcesses
    _PasscryptSettings: type = PasscryptSettings

    # An operator of a passphrase database may add a static secret value
    # to the class, referred to as a `pepper`. That value can be set in
    # this variable & will then augment all hashes produced by the class
    # with that additional secret entropy. The operator is in charge of
    # storing this value securely so it can be reused when the program
    # restarts. This value SHOULD NOT be stored in the same database
    # where the hashes are stored.
    PEPPER: bytes = b""

    TimestampExpired = TimestampExpired
    InvalidPassphrase = InvalidPassphrase
    ImproperPassphrase = ImproperPassphrase

    def __init__(
        self,
        *,
        tag_size: int,
        mb: int = passcrypt_spec.DEFAULT_MB,
        cpu: int = passcrypt_spec.DEFAULT_CPU,
        cores: int = passcrypt_spec.DEFAULT_CORES,
        salt_size: int = passcrypt_spec.DEFAULT_SCHEMA_SALT_SIZE,
        config: t.ConfigType = passcrypt_spec,
    ) -> None:
        """
        Stores user-defined settings.
        """
        self._config = config
        self._settings = self._PasscryptSettings(
            mb=mb,
            cpu=cpu,
            cores=cores,
            tag_size=tag_size,
            salt_size=salt_size,
            config=config,
        )

    async def anew(
        self, passphrase: bytes, salt: bytes, *, aad: bytes = DEFAULT_AAD
    ) -> bytes:
        """
        Returns just the `tag_size`-byte hash of the `passphrase`
        when processed with the given `salt`, `aad` & difficulty
        settings.

        NOTICE: The passcrypt algorithm can be highly memory intensive.
        These resources may not be freed up, & often are not, because of
        python quirks around memory management. To force the release of
        these resources, we run the function in another process which
        guarantees the release.
        """
        try:
            sessions = await self._PasscryptProcesses().aspawn(
                passphrase=passphrase,
                salt=salt,
                aad=aad,
                pepper=self.PEPPER,
                mb=self._settings.mb,
                cpu=self._settings.cpu,
                cores=self._settings.cores,
                tag_size=self._settings.tag_size,
                config=self._config,
            )
            return await ahash_bytes(
                *[await session.aresult() for session in sessions],
                size=self._settings.tag_size,
                hasher=shake_128,
                key=await sessions[0].aresult(),
            )
        except Exception as error:
            if Processes._pool._broken:
                Processes.reset_pool()
                raise Issue.broken_pool_restarted() from error
            raise error

    def new(
        self, passphrase: bytes, salt: bytes, *, aad: bytes = DEFAULT_AAD
    ) -> bytes:
        """
        Returns just the `tag_size`-byte hash of the `passphrase`
        when processed with the given `salt`, `aad` & difficulty
        settings.

        NOTICE: The passcrypt algorithm can be highly memory intensive.
        These resources may not be freed up, & often are not, because of
        python quirks around memory management. To force the release of
        these resources, we run the function in another process which
        guarantees the release.
        """
        try:
            sessions = self._PasscryptProcesses().spawn(
                passphrase=passphrase,
                salt=salt,
                aad=aad,
                pepper=self.PEPPER,
                mb=self._settings.mb,
                cpu=self._settings.cpu,
                cores=self._settings.cores,
                tag_size=self._settings.tag_size,
                config=self._config,
            )
            return hash_bytes(
                *[session.result() for session in sessions],
                size=self._settings.tag_size,
                hasher=shake_128,
                key=sessions[0].result(),
            )
        except Exception as error:
            if Processes._pool._broken:
                Processes.reset_pool()
                raise Issue.broken_pool_restarted() from error
            raise error

    async def ahash_passphrase(
        self, passphrase: bytes, *, aad: bytes = DEFAULT_AAD
    ) -> bytes:
        """
        Returns the passcrypt difficulty settings, salt & hash of the
        `passphrase` in a single raw bytes sequence for convenient
        storage. The salt here is automatically generated.

        `passphrase`: A 12-byte or greater entropic value that
                contains the user's desired entropy & cryptographic
                strength. A passphrase is ideally a compilation of
                words, symbols & numbers which are easy for the user to
                remember, but difficult for anyone else to guess even if
                they know precise details about who the user is & their
                characteristics / tendencies.

        `aad`: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different `aad`.

        NOTICE: The passcrypt algorithm can be highly memory intensive.
        These resources may not be freed up, & often are not, because of
        python quirks around memory management. To force the release of
        these resources, we run the function in another process which
        guarantees the release.
        """
        timestamp = await self._config.clock.amake_timestamp()
        salt = await acsprng(self._settings.salt_size)
        tag = await self.anew(passphrase, salt, aad=timestamp + aad)
        return PasscryptHash(
            timestamp=timestamp,
            mb=self._settings.mb,
            cpu=self._settings.cpu,
            cores=self._settings.cores,
            salt=salt,
            tag=tag,
            config=self._config,
        ).export_hash()

    def hash_passphrase(
        self, passphrase: bytes, *, aad: bytes = DEFAULT_AAD
    ) -> bytes:
        """
        Returns the passcrypt difficulty settings, salt & hash of the
        `passphrase` in a single raw bytes sequence for convenient
        storage. The salt here is automatically generated.

        `passphrase`: A 12-byte or greater entropic value that
                contains the user's desired entropy & cryptographic
                strength. A passphrase is ideally a compilation of
                words, symbols & numbers which are easy for the user to
                remember, but difficult for anyone else to guess even if
                they know precise details about who the user is & their
                characteristics / tendencies.

        `aad`: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different `aad`.

        NOTICE: The passcrypt algorithm can be highly memory intensive.
        These resources may not be freed up, & often are not, because of
        python quirks around memory management. To force the release of
        these resources, we run the function in another process which
        guarantees the release.
        """
        timestamp = self._config.clock.make_timestamp()
        salt = csprng(self._settings.salt_size)
        tag = self.new(passphrase, salt, aad=timestamp + aad)
        return PasscryptHash(
            timestamp=timestamp,
            mb=self._settings.mb,
            cpu=self._settings.cpu,
            cores=self._settings.cores,
            salt=salt,
            tag=tag,
            config=self._config,
        ).export_hash()

    @classmethod
    async def averify(
        cls,
        composed_passcrypt_hash: bytes,
        passphrase: bytes,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
        mb_allowed: range = passcrypt_spec.MB_RESOURCE_SAFETY_RANGE,
        cpu_allowed: range = passcrypt_spec.CPU_RESOURCE_SAFETY_RANGE,
        cores_allowed: range = passcrypt_spec.CORES_RESOURCE_SAFETY_RANGE,
        config: t.ConfigType = passcrypt_spec,
    ) -> None:
        """
        Verifies that a supplied `passphrase` was indeed used to build
        the `composed_passcrypt_hash`.

        `aad`: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different `aad`.

        `ttl`: An amount of seconds which dictate the allowable age of
                a `composed_passcrypt_hash`. The associated timestamp,
                which is attached to the hash, helps ensure the tag is
                unique by separating each tag created across time into
                distinct domains.

        `mb_allowed`: A `builtins.range` object which includes all
                allowable values for the `mb` (Mebibyte) resource cost.
                Raises `ResourceWarning` if the `mb` specified in the
                provided hash falls outside of that range.

        `cpu_allowed`: A `builtins.range` object which includes all
                allowable values for the `cpu` resource cost. Raises
                `ResourceWarning` if the `cpu` specified in the provided
                hash falls outside of that range.

        `cores_allowed`: A `builtins.range` object which includes all
                allowable values for the `cores` resource cost. Raises
                `ResourceWarning` if the `cores` specified in the
                provided hash falls outside of that range.
        """
        parts = PasscryptHash(config=config).import_hash(
            composed_passcrypt_hash
        )
        ttl = ttl if ttl is None else (ttl * NS_TO_S_RATIO)
        await config.clock.atest_timestamp(parts.timestamp, ttl=ttl)
        parts.in_allowed_ranges(mb_allowed, cpu_allowed, cores_allowed)
        self = cls(
            mb=parts.mb,
            cpu=parts.cpu,
            cores=parts.cores,
            tag_size=parts.tag_size,
            config=config,
        )
        untrusted_hash = await self.anew(
            passphrase, parts.salt, aad=parts.timestamp + aad
        )
        if not bytes_are_equal(untrusted_hash, parts.tag):
            raise PasscryptIssue.verification_failed()

    @classmethod
    def verify(
        cls,
        composed_passcrypt_hash: bytes,
        passphrase: bytes,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
        mb_allowed: range = passcrypt_spec.MB_RESOURCE_SAFETY_RANGE,
        cpu_allowed: range = passcrypt_spec.CPU_RESOURCE_SAFETY_RANGE,
        cores_allowed: range = passcrypt_spec.CORES_RESOURCE_SAFETY_RANGE,
        config: t.ConfigType = passcrypt_spec,
    ) -> None:
        """
        Verifies that a supplied `passphrase` was indeed used to build
        the `composed_passcrypt_hash`.

        `aad`: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different `aad`.

        `ttl`: An amount of seconds which dictate the allowable age of
                a `composed_passcrypt_hash`. The associated timestamp,
                which is attached to the hash, helps ensure the tag is
                unique by separating each tag created across time into
                distinct domains.

        `mb_allowed`: A `builtins.range` object which includes all
                allowable values for the `mb` (Mebibyte) resource cost.
                Raises `ResourceWarning` if the `mb` specified in the
                provided hash falls outside of that range.

        `cpu_allowed`: A `builtins.range` object which includes all
                allowable values for the `cpu` resource cost. Raises
                `ResourceWarning` if the `cpu` specified in the provided
                hash falls outside of that range.

        `cores_allowed`: A `builtins.range` object which includes all
                allowable values for the `cores` resource cost. Raises
                `ResourceWarning` if the `cores` specified in the
                provided hash falls outside of that range.
        """
        parts = PasscryptHash(config=config).import_hash(
            composed_passcrypt_hash
        )
        ttl = ttl if ttl is None else (ttl * NS_TO_S_RATIO)
        config.clock.test_timestamp(parts.timestamp, ttl=ttl)
        parts.in_allowed_ranges(mb_allowed, cpu_allowed, cores_allowed)
        self = cls(
            mb=parts.mb,
            cpu=parts.cpu,
            cores=parts.cores,
            tag_size=parts.tag_size,
            config=config,
        )
        untrusted_hash = self.new(
            passphrase, parts.salt, aad=parts.timestamp + aad
        )
        if not bytes_are_equal(untrusted_hash, parts.tag):
            raise PasscryptIssue.verification_failed()


module_api = dict(
    PasscryptSettings=t.add_type(PasscryptSettings),
    Passcrypt=t.add_type(Passcrypt),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Definitions for compossible types to run the dual-output cipher
mode of operation.
"""

__all__ = [
    "DualOutputKDFs",
    "DualOutputSessionKDFs",
    "DualOutputKeyAADBundle",
    "DualOutputStreamHMAC",
    "DualOutputSyntheticIV",
    "DualOutputStreamJunction",
]


from aiootp._typing import Typing as t
from aiootp._constants.misc import BIG
from aiootp._exceptions import Issue
from aiootp.asynchs import asleep
from aiootp.commons import FrozenInstance

from .cipher_kdfs import CipherKDFs
from .key_bundle import KeyAADBundle
from .stream_hmac import StreamHMAC
from .synthetic_iv import SyntheticIV
from .stream_junction import StreamJunction
from .chunky_2048_config import SHMAC_KDF, LEFT_KDF, RIGHT_KDF


class DualOutputKDFs(CipherKDFs):
    """
    A type responsible for initializing the key-derivation & mac objects
    for dual-output mode ciphers.
    """

    __slots__ = (
        "keyed_shmac_kdf",
        "keyed_left_kdf",
        "keyed_right_kdf",
    )

    def __init__(self, key: bytes, *, config: t.ConfigType) -> None:
        """
        Keys the KDFs for handling by `KeyAADBundle`.
        """
        self.config = config
        self.test_key_validity(key)
        self.keyed_shmac_kdf = self.key_base_kdf(SHMAC_KDF, key=key)
        self.keyed_left_kdf = self.key_base_kdf(LEFT_KDF, key=key)
        self.keyed_right_kdf = self.key_base_kdf(RIGHT_KDF, key=key)

    def new_session(self, summary: bytes) -> t.Iterable[t.XOFType]:
        """
        Yields copies of the instance's KDFs that have been given the
        fresh randomness & context of a session's `summary`.
        """
        yield SHMAC_KDF, self.keyed_shmac_kdf.new_session_copy(summary)
        yield LEFT_KDF, self.keyed_left_kdf.new_session_copy(summary)
        yield RIGHT_KDF, self.keyed_right_kdf.new_session_copy(summary)


class DualOutputSessionKDFs(FrozenInstance):
    """
    Efficiently stores & gives access to keyed & session-randomized /
    contextualized functions.
    """

    __slots__ = (
        "keystream",
        "shmac_kdf",
        "left_kdf",
        "right_kdf",
    )

    def __init__(self) -> None:
        pass

    def __iter__(self) -> t.Generator[t.XOFType, None, None]:
        """
        An api for retrieving the instance's keystream kdfs.
        """
        yield self.left_kdf
        yield self.right_kdf


class DualOutputKeyAADBundle(KeyAADBundle):
    """
    A low-level interface for managing a key, salt, iv & authenticated
    associated data bundle which is to be used for ONLY ONE encryption.
    """

    __slots__ = ()

    _Session: type = DualOutputSessionKDFs

    def _keystream_ratchets(
        self,
    ) -> t.Tuple[
        t.Callable[[bytes], None],
        t.Callable[[int], bytes],
        t.Callable[[bytes], None],
        t.Callable[[int], bytes],
    ]:
        """
        Returns the method pointers to the `hashlib.shake_128` objects
        that have been primed in different ways with cipher-specific
        static values, as well as the `key`, `salt`, `aad` & `iv`.

        The pointers are used to construct a key ratchet algorithm.
        """
        left_kdf, right_kdf = self._session
        return (
            left_kdf.update,
            left_kdf.digest,
            right_kdf.update,
            right_kdf.digest,
        )

    async def _anew_keystream(self) -> t.AsyncGenerator[bytes, bytes]:
        """
        An efficient async coroutine producing an unending, non-repeating
        stream of bytes key material which incorporates new key material
        in the stream derivation on each iteration.
        """
        # fmt: off
        config = self.config
        l_update, l_digest, r_update, r_digest = self._keystream_ratchets()
        LEFT_RATCHET_KEY_SLICE, RIGHT_RATCHET_KEY_SLICE, SHMAC_BLOCKSIZE = (
            config.LEFT_RATCHET_KEY_SLICE,
            config.RIGHT_RATCHET_KEY_SLICE,
            config.SHMAC_BLOCKSIZE,
        )
        ratchet_key = yield
        while True:
            l_update(ratchet_key[LEFT_RATCHET_KEY_SLICE])   # update with 168 even index bytes
            r_update(ratchet_key[RIGHT_RATCHET_KEY_SLICE])  # update with 168 odd index bytes
            ratchet_key = yield (
                l_digest(SHMAC_BLOCKSIZE) + r_digest(SHMAC_BLOCKSIZE)
            )
            await asleep()
        # fmt: on

    def _new_keystream(self) -> t.Generator[bytes, bytes, None]:
        """
        An efficient sync coroutine producing an unending, non-repeating
        stream of bytes key material which incorporates new key material
        in the stream derivation on each iteration.
        """
        # fmt: off
        config = self.config
        l_update, l_digest, r_update, r_digest = self._keystream_ratchets()
        LEFT_RATCHET_KEY_SLICE, RIGHT_RATCHET_KEY_SLICE, SHMAC_BLOCKSIZE = (
            config.LEFT_RATCHET_KEY_SLICE,
            config.RIGHT_RATCHET_KEY_SLICE,
            config.SHMAC_BLOCKSIZE,
        )
        ratchet_key = yield
        while True:
            l_update(ratchet_key[LEFT_RATCHET_KEY_SLICE])   # update with 168 even index bytes
            r_update(ratchet_key[RIGHT_RATCHET_KEY_SLICE])  # update with 168 odd index bytes
            ratchet_key = yield (
                l_digest(SHMAC_BLOCKSIZE) + r_digest(SHMAC_BLOCKSIZE)
            )
        # fmt: on

    async def async_mode(self) -> t.Self:
        """
        Sets the instance to run async dual-output mode key derivation.
        """
        self._session.keystream = keystream = self._anew_keystream()
        await keystream.asend(None)
        self._mode.set_async_mode()
        return self

    def sync_mode(self) -> t.Self:
        """
        Sets the instance to run sync dual-output mode key derivation.
        """
        self._session.keystream = keystream = self._new_keystream()
        keystream.send(None)
        self._mode.set_sync_mode()
        return self

    @property
    def _keystream(
        self,
    ) -> t.Union[
        t.Generator[bytes, bytes, None], t.AsyncGenerator[bytes, bytes]
    ]:
        """
        Returns the private keystream coroutine used in the `DualOutput`
        cipher to encrypt / decrypt data. The coroutine can be either
        async or sync depending on what mode the instance is set to.
        """
        return self._session.keystream


class DualOutputStreamHMAC(StreamHMAC):
    """
    This class is used as an inline validator for ciphertext streams as
    they are being created & decrypted. It's designed for full context
    commitment, salt misuse-reuse resistance, & RUP security.
    """

    __slots__ = ()

    @property
    def _ratchet_key(self) -> bytes:
        """
        Combines the current & previous cached state to efficiently
        incorporate fresh entropy from the message into keystream
        generation, as well as prior & current context as commitment.
        This also acts to mitigate adversarial attempts to control the
        internal state of the cipher by spreading/dividing out message
        influence over serveral blocks/outputs of the XOF objects.
        """
        return self._current_digest + self._previous_digest

    async def _aencipher_then_hash(
        self,
        plaintext_block: bytes,
        key: bytes,
        *,
        _from_bytes: t.Callable[..., int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_avalidated_transform`
        method after the encryption mode is set.
        """
        config = self.config
        try:
            ciphertext_block = (
                _from_bytes(plaintext_block, BIG)
                ^ _from_bytes(key[config.EMBEDDED_CIPHERTEXT_SLICE], BIG)
            ).to_bytes(config.BLOCKSIZE, BIG)
            await self._aupdate(
                key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
                + ciphertext_block
                + key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
            )
            return ciphertext_block
        except OverflowError as error:
            raise Issue.exceeded_blocksize(config.BLOCKSIZE) from error

    def _encipher_then_hash(
        self,
        plaintext_block: bytes,
        key: bytes,
        *,
        _from_bytes: t.Callable[..., int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_validated_transform`
        method after the encryption mode is set.
        """
        config = self.config
        try:
            ciphertext_block = (
                _from_bytes(plaintext_block, BIG)
                ^ _from_bytes(key[config.EMBEDDED_CIPHERTEXT_SLICE], BIG)
            ).to_bytes(config.BLOCKSIZE, BIG)
            self._update(
                key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
                + ciphertext_block
                + key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
            )
            return ciphertext_block
        except OverflowError as error:
            raise Issue.exceeded_blocksize(config.BLOCKSIZE) from error

    async def _ahash_then_decipher(
        self,
        ciphertext_block: bytes,
        key: bytes,
        *,
        _from_bytes: t.Callable[..., int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_avalidated_transform`
        method after the decryption mode is set.
        """
        config = self.config
        try:
            await self._aupdate(
                key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
                + ciphertext_block
                + key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
            )
            return (
                _from_bytes(ciphertext_block, BIG)
                ^ _from_bytes(key[config.EMBEDDED_CIPHERTEXT_SLICE], BIG)
            ).to_bytes(config.BLOCKSIZE, BIG)
        except OverflowError as error:
            raise Issue.exceeded_blocksize(config.BLOCKSIZE) from error

    def _hash_then_decipher(
        self,
        ciphertext_block: bytes,
        key: bytes,
        *,
        _from_bytes: t.Callable[..., int] = int.from_bytes,
    ) -> bytes:
        """
        This method is inserted as the instance's `_validated_transform`
        method after the decryption mode is set.
        """
        config = self.config
        try:
            self._update(
                key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
                + ciphertext_block
                + key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
            )
            return (
                _from_bytes(ciphertext_block, BIG)
                ^ _from_bytes(key[config.EMBEDDED_CIPHERTEXT_SLICE], BIG)
            ).to_bytes(config.BLOCKSIZE, BIG)
        except OverflowError as error:
            raise Issue.exceeded_blocksize(config.BLOCKSIZE) from error


class DualOutputSyntheticIV(SyntheticIV):
    """
    Manages the application of synthetic IVs which improve the salt
    misuse-reuse resistance of dual-output mode ciphers, since if either
    the 4-byte timestamp or ephemeral SIV-key are unique, then the
    entire stream of key material will be unique. The required plaintext
    padding is handled within the `Padding` class.

     _____________________________________
    |                                     |
    |    Algorithm Diagram: Encryption    |
    |_____________________________________|

     ------------------------------------------------------------------
    |    H = inner-header     |      P = first block of plaintext      |
    |                         |                                        |
    |         I-bytes         |                D-bytes                 |
    |-------------------------|----------------------------------------|
    |  timestamp |  siv-key   |                                        |
     ------------------------------------------------------------------

    |---------------------------- B-bytes -----------------------------|
    |----------- H -----------|------------------- P ------------------|
                 |                                 |
                 v                                 |
     shmac.perm( H )                               |
         |                                         |
         v                                         |
    M = masked-inner-header                        |
    J = (336 + I - B) / 2                          |
    L = (336 - B) / 2                              |
    siv = H + M + shmac.digest(168 - len(H + M))   |
    shmac.update(siv)                              |
    key = keystream(shmac.digest(336))             |
                                                   |
    key[J:-J] -------------------------------------⊕----> C
                                                   |
    shmac.update(key[:L] + M + C + key[-L:])       |
                                                   |
                                                   v
     ------------------------------------------------------------------
    | M = masked-inner-header |     C = first block of ciphertext      |
     ------------------------------------------------------------------

     _____________________________________
    |                                     |
    |    Algorithm Diagram: Decryption    |
    |_____________________________________|

     ------------------------------------------------------------------
    | M = masked-inner-header |     C = first block of ciphertext      |
     ------------------------------------------------------------------

    |---------------------------- B-bytes -----------------------------|
    |----------- M -----------|------------------- C ------------------|
                 |                                 |
                 v                                 |
     shmac.perm( M )                               |
         |                                         |
         v                                         |
    H = inner-header                               |
    siv = H + M + shmac.digest(168 - len(H + M))   |
    shmac.update(siv)                              |
    key = keystream(shmac.digest(336))             |
                                                   |
    key[J:-J] -------------------------------------⊕----> P
                                                   |
    shmac.update(key[:L] + M + C + key[-L:])       |
                                                   |
                                                   v
     ------------------------------------------------------------------
    |    H = inner-header     |      P = first block of plaintext      |
    |                         |                                        |
    |         I-bytes         |                D-bytes                 |
    |-------------------------|----------------------------------------|
    |  timestamp |  siv-key   |                                        |
     ------------------------------------------------------------------
    """

    __slots__ = ()

    @classmethod
    async def _aunique_cipher(
        cls,
        block: bytes,
        keystream: t.Callable[[bytes], t.Awaitable[bytes]],
        shmac: t.StreamHMACType,
    ) -> bytes:
        """
        Uses a masking & encryption algorithm to pull secret ephemeral
        data from the plaintext header to protect the payload of the
        ciphertext with more salt misuse-reuse resistance.
        """
        config = shmac.config
        header = block[config.INNER_HEADER_SLICE]
        masked_header = (
            await shmac._permutation.apermute(int.from_bytes(header, BIG))
        ).to_bytes(config.INNER_HEADER_BYTES, BIG)
        shmac._mac.update(
            header
            + masked_header
            + shmac._current_digest[config.SIV_DIGEST_SLICE]
        )
        key = await keystream(
            shmac._mac.digest(config.SHMAC_DOUBLE_BLOCKSIZE)
        )
        l_capacity = key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
        r_capacity = key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
        ciphertext = masked_header + (
            int.from_bytes(key[config.FIRST_KEY_SLICE], BIG)
            ^ int.from_bytes(block[config.FIRST_CONTENT_SLICE], BIG)
        ).to_bytes(config.FIRST_CONTENT_BYTES, BIG)
        await shmac._aupdate(l_capacity + ciphertext + r_capacity)
        return ciphertext

    @classmethod
    def _unique_cipher(
        cls,
        block: bytes,
        keystream: t.Callable[[bytes], bytes],
        shmac: t.StreamHMACType,
    ) -> bytes:
        """
        Uses a masking & encryption algorithm to pull secret ephemeral
        data from the plaintext header to protect the payload of the
        ciphertext with more salt misuse-reuse resistance.
        """
        config = shmac.config
        header = block[config.INNER_HEADER_SLICE]
        masked_header = shmac._permutation.permute(
            int.from_bytes(header, BIG)
        ).to_bytes(config.INNER_HEADER_BYTES, BIG)
        shmac._mac.update(
            header
            + masked_header
            + shmac._current_digest[config.SIV_DIGEST_SLICE]
        )
        key = keystream(shmac._mac.digest(config.SHMAC_DOUBLE_BLOCKSIZE))
        l_capacity = key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
        r_capacity = key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
        ciphertext = masked_header + (
            int.from_bytes(key[config.FIRST_KEY_SLICE], BIG)
            ^ int.from_bytes(block[config.FIRST_CONTENT_SLICE], BIG)
        ).to_bytes(config.FIRST_CONTENT_BYTES, BIG)
        shmac._update(l_capacity + ciphertext + r_capacity)
        return ciphertext

    @classmethod
    async def _aunique_decipher(
        cls,
        block: bytes,
        keystream: t.Callable[[bytes], t.Awaitable[bytes]],
        shmac: t.StreamHMACType,
    ) -> bytes:
        """
        Uses an unmasking & decryption algorithm to pull secret
        ephemeral data from the plaintext header to protect the payload
        of the ciphertext with more salt misuse-reuse resistance.
        """
        config = shmac.config
        masked_header = block[config.INNER_HEADER_SLICE]
        header = (
            await shmac._permutation.ainvert(
                int.from_bytes(masked_header, BIG)
            )
        ).to_bytes(config.INNER_HEADER_BYTES, BIG)
        shmac._mac.update(
            header
            + masked_header
            + shmac._current_digest[config.SIV_DIGEST_SLICE]
        )
        key = await keystream(
            shmac._mac.digest(config.SHMAC_DOUBLE_BLOCKSIZE)
        )
        l_capacity = key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
        r_capacity = key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
        plaintext = header + (
            int.from_bytes(key[config.FIRST_KEY_SLICE], BIG)
            ^ int.from_bytes(block[config.FIRST_CONTENT_SLICE], BIG)
        ).to_bytes(config.FIRST_CONTENT_BYTES, BIG)
        await shmac._aupdate(l_capacity + block + r_capacity)
        return plaintext

    @classmethod
    def _unique_decipher(
        cls,
        block: bytes,
        keystream: t.Callable[[bytes], bytes],
        shmac: t.StreamHMACType,
    ) -> bytes:
        """
        Uses an unmasking & decryption algorithm to pull secret
        ephemeral data from the plaintext header to protect the payload
        of the ciphertext with more salt misuse-reuse resistance.
        """
        config = shmac.config
        masked_header = block[config.INNER_HEADER_SLICE]
        header = shmac._permutation.invert(
            int.from_bytes(masked_header, BIG)
        ).to_bytes(config.INNER_HEADER_BYTES, BIG)
        shmac._mac.update(
            header
            + masked_header
            + shmac._current_digest[config.SIV_DIGEST_SLICE]
        )
        key = keystream(shmac._mac.digest(config.SHMAC_DOUBLE_BLOCKSIZE))
        l_capacity = key[config.EMBEDDED_LEFT_CAPACITY_SLICE]
        r_capacity = key[config.EMBEDDED_RIGHT_CAPACITY_SLICE]
        plaintext = header + (
            int.from_bytes(key[config.FIRST_KEY_SLICE], BIG)
            ^ int.from_bytes(block[config.FIRST_CONTENT_SLICE], BIG)
        ).to_bytes(config.FIRST_CONTENT_BYTES, BIG)
        shmac._update(l_capacity + block + r_capacity)
        return plaintext


class DualOutputStreamJunction(StreamJunction):
    """
    A definition for how the key & data streams are combined for the
    dual-output cipher modes.
    """

    __slots__ = ()

    _SyntheticIV: type = DualOutputSyntheticIV

    @classmethod
    async def acombine_streams(
        cls, data: t.AsyncDatastream, *, shmac: t.StreamHMACType
    ) -> t.AsyncGenerator[bytes, None]:
        """
        Bitwise XORs a ciphertext or plaintext datastream with a keystream
        then feeds ciphertext into the `shmac` for validation & distinct
        altering of the keystream.
        """
        datastream, keystream, validated_transform = (
            data,
            shmac._key_bundle._keystream.asend,
            shmac._avalidated_transform,
        )
        yield await cls._SyntheticIV.avalidated_transform(
            datastream, keystream, shmac
        )
        async for block in datastream:
            yield await validated_transform(
                block, await keystream(shmac._ratchet_key)
            )

    @classmethod
    def combine_streams(
        cls, data: t.Datastream, *, shmac: t.StreamHMACType
    ) -> t.Generator[bytes, None, None]:
        """
        Bitwise XORs a ciphertext or plaintext datastream with a keystream
        then feeds ciphertext into the `shmac` for validation & distinct
        altering of the keystream.
        """
        datastream, keystream, validated_transform = (
            data,
            shmac._key_bundle._keystream.send,
            shmac._validated_transform,
        )
        yield cls._SyntheticIV.validated_transform(
            datastream, keystream, shmac
        )
        for block in datastream:
            yield validated_transform(block, keystream(shmac._ratchet_key))


module_api = dict(
    DualOutputKDFs=t.add_type(DualOutputKDFs),
    DualOutputSessionKDFs=t.add_type(DualOutputSessionKDFs),
    DualOutputKeyAADBundle=t.add_type(DualOutputKeyAADBundle),
    DualOutputStreamHMAC=t.add_type(DualOutputStreamHMAC),
    DualOutputSyntheticIV=t.add_type(DualOutputSyntheticIV),
    DualOutputStreamJunction=t.add_type(DualOutputStreamJunction),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)
